import base64

import requests
from steamship_tests import APPS_PATH, TEST_ASSETS_PATH
from steamship_tests.utils.deployables import deploy_app
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import Space
from steamship.base import TaskState
from steamship.base.mime_types import MimeTypes
from steamship.data.user import User


def _fix_url(s: str) -> str:
    """Homogenize references to `this machine` for the purpose of comparing remote configuration and local
    configuration. The goal of the below steamship_tests isn't to check that your machine has been configured in the
    "approved way" (which is to use host.docker.internal). It is merely to make sure that the environment
    has been passed to the app instance correctly."""
    s = s.replace("localhost", "host.docker.internal").replace("127.0.0.1", "host.docker.internal")
    if s.endswith("/"):
        s = s[:-1]  # s.removesuffix is only available in Python 3.9; we use Python 3.8
    return s


def test_instance_invoke():
    palm_tree_path = TEST_ASSETS_PATH / "palm_tree.png"

    with palm_tree_path.open("rb") as f:
        palm_bytes = f.read()
    base64_palm = base64.b64encode(palm_bytes).decode("utf-8")

    client = get_steamship_client()
    demo_app_path = APPS_PATH / "demo_app.py"

    with deploy_app(client, demo_app_path) as (app, version, instance):
        # Now let's invoke it!
        # Note: we're invoking the data at demo_app.py in the steamship_tests/demo_apps folder

        def get_raw(path: str):
            return requests.get(
                instance.full_url_for(path),
                headers={"authorization": f"Bearer {client.config.api_key}"},
            )

        res = instance.get("greet").data
        assert res == "Hello, Person!"

        resp = get_raw("greet")
        assert resp.text == "Hello, Person!"

        res = instance.get("greet", name="Ted").data
        assert res == "Hello, Ted!"
        url = instance.full_url_for("greet?name=Ted")
        resp = requests.get(url, headers={"authorization": f"Bearer {client.config.api_key}"})
        assert resp.text == "Hello, Ted!"

        res = instance.post("greet").data
        assert res == "Hello, Person!"
        url = instance.full_url_for("greet")
        resp = requests.post(url, headers={"authorization": f"Bearer {client.config.api_key}"})
        assert resp.text == "Hello, Person!"

        res = instance.post("greet", name="Ted").data
        assert res == "Hello, Ted!"
        url = instance.full_url_for("greet")
        resp = requests.post(
            url,
            json={"name": "Ted"},
            headers={"authorization": f"Bearer {client.config.api_key}"},
        )
        assert resp.text == "Hello, Ted!"

        # Now we test different return types
        resp_string = get_raw("resp_string")
        assert resp_string.text == "A String"

        resp_dict = get_raw("resp_dict")
        assert resp_dict.json() == {"string": "A String", "int": 10}

        resp_404 = get_raw("doesnt_exist")
        json_404 = resp_404.json()
        assert isinstance(json_404, dict)
        assert json_404.get("status") is not None
        assert json_404.get("status") is not None
        assert json_404.get("status", {}).get("state") == TaskState.failed
        # assert "No handler" in json_404.get("status", dict()).get("statusMessage", "")
        assert resp_404.status_code == 404

        resp_obj = get_raw("resp_obj")
        assert resp_obj.json() == {"name": "Foo"}

        resp_binary = get_raw("resp_binary")
        base64_binary = base64.b64encode(resp_binary.content).decode("utf-8")
        assert base64_binary == base64_palm
        assert resp_binary.headers.get("Content-Type") == MimeTypes.BINARY

        resp_bytes_io = get_raw("resp_bytes_io")
        base64_bytes_io = base64.b64encode(resp_bytes_io.content).decode("utf-8")
        assert base64_bytes_io == base64_palm
        assert resp_bytes_io.headers.get("Content-Type") == MimeTypes.BINARY

        resp_image = get_raw("resp_image")
        base64_image = base64.b64encode(resp_image.content).decode("utf-8")
        assert base64_image == base64_palm
        assert resp_image.headers.get("Content-Type") == MimeTypes.PNG

        # The test app, when executing remotely inside Steamship, should have the same
        # set of configuration options that we're running with here within the test
        configuration_within_lambda_resp = instance.get("config")
        configuration_within_lambda = configuration_within_lambda_resp.data

        my_app_base = _fix_url(client.config.app_base)
        remote_app_base = _fix_url(configuration_within_lambda["appBase"])

        my_api_base = _fix_url(client.config.api_base)
        remote_api_base = _fix_url(configuration_within_lambda["apiBase"])

        assert my_app_base == remote_app_base
        assert my_api_base == remote_api_base
        assert configuration_within_lambda["apiKey"] == client.config.api_key

        # SpaceId is an exception. Rather than being the SpaceId of the client, it should be the SpaceId
        # of the App Instance.
        assert configuration_within_lambda["spaceId"] == instance.space_id  # SpaceID

        # The test app, when executing remotely inside Steamship, should have the same
        # user that we're running with here in within the test
        user_info = instance.post("user_info")
        user = User.current(client)
        assert user_info.data["handle"] == user.data.handle


def test_deploy_in_space():
    client = get_steamship_client()
    demo_app_path = APPS_PATH / "demo_app.py"

    space = Space.create(client, handle="test-non-default-space").data
    with deploy_app(client, demo_app_path, space_id=space.id) as (_, _, instance):
        assert instance.space_id == space.id



from __future__ import annotations

import time
from typing import Any, Generic, Type, TypeVar

from pydantic.generics import GenericModel

from steamship.base.error import SteamshipError
from steamship.base.tasks import Task, TaskState, TaskStatusRequest
from steamship.base.utils import to_camel

T = TypeVar("T")  # Declare type variable


class Response(GenericModel, Generic[T]):
    expect: Type[T] = None
    task: Task = None
    data_: T = None
    error: SteamshipError = None
    client: Any = None

    class Config:
        arbitrary_types_allowed = True  # This is required to support SteamshipError
        alias_generator = to_camel
        allow_population_by_field_name = True

    @property
    def data(self):
        if self.error:
            raise self.error
        if self.task is not None:
            if self.task.state == TaskState.failed:
                raise SteamshipError(
                    "Asynchronous call failed. No error available. Please contact support."
                )
            if self.task.state != TaskState.succeeded and self.data_ is None:
                raise SteamshipError(
                    "Data not available yet for asynchronous call. Please call .wait() before accessing .data"
                )
        return self.data_

    def update(self, response: Response[T]):
        if self.task is not None and response.task is not None:
            self.task.update(response.task)
        if response.data_ is not None:
            self.data_ = response.data_
        self.error = response.error

    def wait(self, max_timeout_s: float = 60, retry_delay_s: float = 1):
        """Polls and blocks until the task has succeeded or failed (or timeout reached)."""
        if self.task is None:
            return
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < max_timeout_s and self.task.state not in (
            TaskState.succeeded,
            TaskState.failed,
        ):
            self.refresh()
            time.sleep(retry_delay_s)

        # If the task did not complete within the timout, throw an error
        if self.task.state not in (TaskState.succeeded, TaskState.failed):
            raise SteamshipError(
                message=f"Task {self.task.task_id} did not complete within requested timeout of {max_timeout_s}s"
            )

    def refresh(self):
        if self.task is not None:
            req = TaskStatusRequest(taskId=self.task.task_id)
            resp = self.client.post("task/status", payload=req, expect=self.expect)
            self.update(resp)



from typing import Type

from steamship import File
from steamship.app import Response, create_handler
from steamship.data.operations.corpus_importer import CorpusImportRequest, CorpusImportResponse
from steamship.plugin.config import Config
from steamship.plugin.corpus_importer import CorpusImporter
from steamship.plugin.service import PluginRequest


class TestCorpusImporterPlugin(CorpusImporter):
    def config_cls(self) -> Type[Config]:
        return Config

    def run(self, request: PluginRequest[CorpusImportRequest]) -> Response[CorpusImportResponse]:
        return Response(
            data=CorpusImportResponse(
                file_import_requests=[
                    File.CreateRequest(
                        type="fileImporter",
                        corpusId=request.data.url,
                        plugin_instance=request.data.file_importer_plugin_instance,
                    ),
                    File.CreateRequest(
                        type="fileImporter",
                        corpusId=request.data.url,
                        plugin_instance=request.data.file_importer_plugin_instance,
                    ),
                ]
            )
        )


handler = create_handler(TestCorpusImporterPlugin)



from __future__ import annotations

from typing import Any, Dict, List, Optional, Type

from pydantic import BaseModel

from steamship.base import Client, Request
from steamship.base.configuration import CamelModel
from steamship.base.response import Response
from steamship.data.plugin import HostingMemory, HostingTimeout


class CreatePluginVersionRequest(Request):
    plugin_id: str = None
    handle: str = None
    upsert: bool = None
    hosting_memory: Optional[HostingMemory] = None
    hosting_timeout: Optional[HostingTimeout] = None
    hosting_handler: str = None
    is_public: bool = None
    is_default: bool = None
    type: str = "file"
    config_template: Dict[str, Any] = None

    def to_dict(self):
        # Note: the to_dict is necessary here to properly serialize the enum values.
        return {
            "pluginId": self.plugin_id,
            "handle": self.handle,
            "upsert": self.upsert,
            "hostingMemory": self.hosting_memory.value if self.hosting_memory else None,
            "hostingTimeout": self.hosting_timeout.value if self.hosting_timeout else None,
            "hostingHandler": self.hosting_handler,
            "isPublic": self.is_public,
            "isDefault": self.is_default,
            "type": self.type,
            "configTemplate": self.config_template,
        }


class DeletePluginVersionRequest(Request):
    id: str


class ListPublicPluginVersionsRequest(Request):
    handle: str
    plugin_id: str


class ListPrivatePluginVersionsRequest(Request):
    handle: str
    plugin_id: str


class ListPluginVersionsResponse(Response):
    plugins: List[PluginVersion]


class PluginVersion(CamelModel):
    client: Client = None
    id: str = None
    plugin_id: str = None
    handle: str = None
    hosting_memory: Optional[HostingMemory] = None
    hosting_timeout: Optional[HostingTimeout] = None
    hosting_handler: str = None
    is_public: bool = None
    is_default: bool = None
    config_template: Dict[str, Any] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["pluginVersion"] if "pluginVersion" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        handle: str,
        plugin_id: str = None,
        filename: str = None,
        filebytes: bytes = None,
        upsert: bool = False,
        hosting_memory: Optional[HostingMemory] = None,
        hosting_timeout: Optional[HostingTimeout] = None,
        hosting_handler: str = None,
        is_public: bool = None,
        is_default: bool = None,
        config_template: Dict[str, Any] = None,
    ) -> Response[PluginVersion]:

        if filename is None and filebytes is None:
            raise Exception("Either filename or filebytes must be provided.")
        if filename is not None and filebytes is not None:
            raise Exception("Only either filename or filebytes should be provided.")

        if filename is not None:
            with open(filename, "rb") as f:
                filebytes = f.read()

        req = CreatePluginVersionRequest(
            handle=handle,
            plugin_id=plugin_id,
            upsert=upsert,
            hosting_memory=hosting_memory,
            hosting_timeout=hosting_timeout,
            hosting_handler=hosting_handler,
            is_public=is_public,
            is_default=is_default,
            config_template=config_template,
        )

        return client.post(
            "plugin/version/create",
            payload=req,
            file=("plugin.zip", filebytes, "multipart/form-data"),
            expect=PluginVersion,
        )

    def delete(self) -> PluginVersion:
        req = DeletePluginVersionRequest(id=self.id)
        return self.client.post("plugin/version/delete", payload=req, expect=PluginVersion)

    @staticmethod
    def get_public(client: Client, plugin_id: str, handle: str):
        public_plugins = PluginVersion.list_public(
            client=client, plugin_id=plugin_id, handle=handle
        ).data.plugins
        for plugin in public_plugins:
            if plugin.handle == handle:
                return plugin
        return None

    @staticmethod
    def list_public(
        client: Client, plugin_id: str = None, handle: str = None
    ) -> Response[ListPluginVersionsResponse]:
        return client.post(
            "plugin/version/public",
            ListPublicPluginVersionsRequest(handle=handle, plugin_id=plugin_id),
            expect=ListPluginVersionsResponse,
        )

    @staticmethod
    def list_private(
        client: Client, _=None, plugin_id: str = None, handle: str = None
    ) -> Response[ListPluginVersionsResponse]:
        return client.post(
            "plugin/version/private",
            ListPrivatePluginVersionsRequest(handle=handle, plugin_id=plugin_id),
            expect=ListPluginVersionsResponse,
        )



from assets.plugins.importers.plugin_corpus_importer import TestCorpusImporterPlugin

from steamship.app import Response
from steamship.data.operations.corpus_importer import CorpusImportRequest, CorpusImportResponse
from steamship.plugin.service import PluginRequest

TEST_REQ = CorpusImportRequest(url="1")
TEST_PLUGIN_REQ = PluginRequest(data=TEST_REQ)
TEST_PLUGIN_REQ_DICT = TEST_PLUGIN_REQ.dict()


def _test_resp(res):
    assert isinstance(res, Response)
    assert isinstance(res.data, CorpusImportResponse)
    assert res.data.file_import_requests is not None
    assert len(res.data.file_import_requests) == 2


def test_importer():
    importer = TestCorpusImporterPlugin()
    res = importer.run(TEST_PLUGIN_REQ)
    _test_resp(res)

    # The endpoints take a kwargs block which is transformed into the appropriate JSON object
    res2 = importer.run_endpoint(**TEST_PLUGIN_REQ_DICT)
    _test_resp(res2)



import logging
from os import environ
from typing import Optional


class Verb:
    GET = "GET"
    POST = "POST"

    @staticmethod
    def safely_from_str(s: str) -> str:
        # TODO (enias): Is this needed?
        ss = s.strip().upper()
        if ss == Verb.GET:
            return Verb.GET
        elif ss == Verb.POST:
            return Verb.POST
        return s


def is_local(base: str) -> bool:
    """Check if we are running the client locally."""
    return any(
        local_base in base
        for local_base in ("localhost", "127.0.0.1", "0:0:0:0", "host.docker.internal", "/test:")
    )


def apply_localstack_url_fix(url: Optional[str]) -> Optional[str]:
    logging.debug(f"URL {url}")
    if url and "host.docker.internal" in url:
        localstack_hostname = environ.get("LOCALSTACK_HOSTNAME")
        # Note: I realize this expression below looks weird, but deleting the left-most predicate
        # cause it to become incorrect.
        if localstack_hostname and localstack_hostname != "localhost":
            logging.info(f"Replacing domain in {url} with {localstack_hostname}")
            return url.replace("host.docker.internal", localstack_hostname)
    return url



from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import temporary_space

from steamship import File
from steamship.data.operations.corpus_importer import CorpusImportRequest, CorpusImportResponse

HANDLE = "test-importer-plugin-v1"
TEST_H1 = "A Poem"
TEST_S1 = "Roses are red."
TEST_S2 = "Violets are blue."
TEST_S3 = "Sugar is sweet, and I love you."
TEST_DOC = f"# {TEST_H1}\n\n{TEST_S1} {TEST_S2}\n\n{TEST_S3}\n"


def test_e2e_corpus_importer():
    client = get_steamship_client()
    corpus_importer_path = PLUGINS_PATH / "importers" / "plugin_corpus_importer.py"
    file_importer_path = PLUGINS_PATH / "importers" / "plugin_file_importer.py"

    with temporary_space(client) as space:
        with deploy_plugin(client, file_importer_path, "fileImporter", space_id=space.id) as (
            _,
            _,
            fi_instance,
        ):
            with deploy_plugin(
                client, corpus_importer_path, "corpusImporter", space_id=space.id
            ) as (
                plugin,
                version,
                instance,
            ):
                req = CorpusImportRequest(
                    type="file",
                    value="dummy-value",
                    plugin_instance=instance.handle,
                    file_importer_plugin_instance=fi_instance.handle,
                )
                res = client.post(
                    "plugin/instance/importCorpus",
                    req,
                    expect=CorpusImportResponse,
                    space_id=space.id,
                )
                res.wait()

                # We should now have two files!
                files = File.list(client, space_id=space.id).data
                assert files.files is not None
                assert len(files.files) == 2

                for file in files.files:
                    data = file.raw().data
                    assert data.decode("utf-8") == TEST_DOC
                    file.delete()



from __future__ import annotations

from typing import Any, Dict, Type

from pydantic import BaseModel

from steamship.base import Client, Request, Response
from steamship.base.configuration import CamelModel
from steamship.data.space import Space


class CreateAppInstanceRequest(Request):
    id: str = None
    app_id: str = None
    app_handle: str = None
    app_version_id: str = None
    app_version_handle: str = None
    handle: str = None
    upsert: bool = None
    config: Dict[str, Any] = None
    space_id: str = None


class DeleteAppInstanceRequest(Request):
    id: str


class AppInstance(CamelModel):
    client: Client = None
    id: str = None
    handle: str = None
    app_id: str = None
    app_handle: str = None
    user_handle: str = None
    app_version_id: str = None
    user_id: str = None
    invocation_url: str = None
    config: Dict[str, Any] = None
    space_id: str = None
    space_handle: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["appInstance"] if "appInstance" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        space_id: str = None,
        app_id: str = None,
        app_handle: str = None,
        app_version_id: str = None,
        app_version_handle: str = None,
        handle: str = None,
        upsert: bool = None,
        config: Dict[str, Any] = None,
    ) -> Response[AppInstance]:
        req = CreateAppInstanceRequest(
            handle=handle,
            app_id=app_id,
            app_handle=app_handle,
            app_version_id=app_version_id,
            app_version_handle=app_version_handle,
            upsert=upsert,
            config=config,
            spaceId=space_id,
        )

        return client.post(
            "app/instance/create", payload=req, expect=AppInstance, space_id=space_id
        )

    def delete(self) -> AppInstance:
        req = DeleteAppInstanceRequest(id=self.id)
        return self.client.post("app/instance/delete", payload=req, expect=AppInstance)

    def load_missing_vals(self):
        if self.client is not None and self.space_handle is None and self.space_id is not None:
            # Get the spaceHandle
            space = Space.get(self.client, id_=self.space_id)
            if space and space.data:
                self.space_handle = space.data.handle

    def get(self, path: str, **kwargs):
        self.load_missing_vals()
        if path[0] == "/":
            path = path[1:]
        return self.client.get(
            f"/{self.space_handle or '_'}/{self.handle or '_'}/{path}",  # TODO (enias): Fix code duplication
            payload=kwargs,
            app_call=True,
            app_owner=self.user_handle,
            app_id=self.app_id,
            app_instance_id=self.id,
            space_id=self.space_id,
        )

    def post(self, path: str, **kwargs):
        self.load_missing_vals()
        if path[0] == "/":
            path = path[1:]
        return self.client.post(
            f"/{self.space_handle or '_'}/{self.handle or '_'}/{path}",
            payload=kwargs,
            app_call=True,
            app_owner=self.user_handle,
            app_id=self.app_id,
            app_instance_id=self.id,
            space_id=self.space_id,
        )

    def full_url_for(self, path: str):
        return f"{self.invocation_url}{path}"


class ListPrivateAppInstancesRequest(Request):
    pass



import contextlib
import random
import string

from steamship import EmbeddingIndex, Space, Steamship


def random_name() -> str:
    letters = string.digits + string.ascii_letters
    return f"test_{''.join(random.choice(letters) for _ in range(10))}"  # noqa: S311


_TEST_EMBEDDER = "test-embedder-v1"


@contextlib.contextmanager
def random_index(steamship: Steamship, plugin_instance: str) -> EmbeddingIndex:
    index = steamship.create_index(plugin_instance=plugin_instance).data
    yield index
    index.delete()  # or whatever you need to do at exit


@contextlib.contextmanager
def temporary_space(steamship: Steamship) -> Space:
    space = Space.create(client=steamship).data
    yield space
    space.delete()  # or whatever you need to do at exit



import re
from typing import Optional

import inflection


def format_uri(uri: Optional[str]) -> Optional[str]:
    if uri is not None and not uri.endswith("/"):
        uri += "/"
    return uri


def to_snake_case(s: str) -> str:
    return inflection.underscore(s)


def to_camel(s: str) -> str:
    s = re.sub("_(url)$", lambda m: f"_{m.group(1).upper()}", s)
    return inflection.camelize(s, uppercase_first_letter=False)



from abc import ABC, abstractmethod
from typing import Any, Dict, Type

from steamship.app import Response, post
from steamship.base import Client
from steamship.plugin.config import Config
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.service import PluginRequest, PluginService

# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a Blockifier.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.converter instead
# of this file.
#


class Blockifier(PluginService[RawDataPluginInput, BlockAndTagPluginOutput], ABC):
    # noinspection PyUnusedLocal
    def __init__(self, client: Client = None, config: Dict[str, Any] = None):
        super().__init__(client, config)
        if config:
            self.config = self.config_cls()(**config)

    @abstractmethod
    def config_cls(self) -> Type[Config]:
        raise NotImplementedError()

    @abstractmethod
    def run(self, request: PluginRequest[RawDataPluginInput]) -> Response[BlockAndTagPluginOutput]:
        raise NotImplementedError()

    @post("blockify")
    def run_endpoint(self, **kwargs) -> Response[BlockAndTagPluginOutput]:
        """Exposes the Corpus Importer's `run` operation to the Steamship Engine via the expected HTTP path POST /import"""
        return self.run(PluginRequest[RawDataPluginInput].parse_obj(kwargs))



from __future__ import annotations

import pytest

from steamship import SteamshipError
from steamship.base import Response, Task, TaskState

TEST_DATA = {"test": "test"}


def test_no_data_response():
    response = Response[dict](error=SteamshipError())

    with pytest.raises(SteamshipError):
        _ = response.data


def test_async_response():
    response = Response[dict](task=Task())
    with pytest.raises(SteamshipError):
        _ = response.data

    response.update(Response[dict](task=Task(state=TaskState.succeeded), data_=TEST_DATA))
    assert response.data is not None
    assert response.data == TEST_DATA

    for ongoing_state in (TaskState.running, TaskState.waiting):
        response = Response[dict](task=Task(state=ongoing_state))
        with pytest.raises(SteamshipError):
            _ = response.data

        response.update(Response[dict](task=Task(state=ongoing_state), data_=TEST_DATA))
        assert response.data is not None
        assert response.data == TEST_DATA

    response = Response[dict](task=Task(state=TaskState.failed))
    with pytest.raises(SteamshipError):
        _ = response.data

    response.update(Response[dict](task=Task(state=TaskState.failed), data_=TEST_DATA))
    with pytest.raises(SteamshipError):
        _ = response.data


def test_data_response():
    response = Response[dict](data_=TEST_DATA)
    assert response.data is not None
    assert response.data == TEST_DATA



from __future__ import annotations

import json
from typing import Any, Dict, List

from steamship.base import Client, Request, Response
from steamship.base.configuration import CamelModel


class TagQueryRequest(Request):
    tag_filter_query: str


class Tag(CamelModel):
    client: Client = None
    id: str = None
    file_id: str = None
    block_id: str = None
    kind: str = None  # E.g. ner
    name: str = None  # E.g. person
    value: Dict[str, Any] = None  # JSON Metadata
    start_idx: int = None  # w/r/t block.text. None means 0 if blockId is not None
    end_idx: int = None  # w/r/t block.text. None means -1 if blockId is not None

    class CreateRequest(Request):
        id: str = None
        file_id: str = None
        block_id: str = None
        kind: str = None
        name: str = None
        start_idx: int = None
        end_idx: int = None
        value: Dict[str, Any] = None
        upsert: bool = None

    class DeleteRequest(Request):
        id: str = None
        file_id: str = None
        block_id: str = None

    class ListRequest(Request):
        file_id: str = None
        block_id: str = None

    class ListResponse(Response):
        tags: List[Tag] = None

    @staticmethod
    def create(
        client: Client,
        file_id: str = None,
        block_id: str = None,
        kind: str = None,
        name: str = None,
        start_idx: int = None,
        end_idx: int = None,
        value: Any = None,
        upsert: bool = None,
        space_id: str = None,
        space_handle: str = None,
    ) -> Response[Tag]:
        if isinstance(value, dict) or isinstance(value, list):
            value = json.dumps(value)

        req = Tag.CreateRequest(
            file_id=file_id,
            block_id=block_id,
            kind=kind,
            name=name,
            start_idx=start_idx,
            end_idx=end_idx,
            value=value,
            upsert=upsert,
        )
        return client.post(
            "tag/create", req, expect=Tag, space_id=space_id, space_handle=space_handle
        )

    @staticmethod
    def list_public(
        client: Client,
        file_id: str = None,
        block_id: str = None,
        space_id: str = None,
        space_handle: str = None,
    ) -> Response[Tag.ListResponse]:
        return client.post(
            "tag/list",
            Tag.ListRequest(file_id=file_id, block_id=block_id),
            expect=Tag.ListResponse,
            space_id=space_id,
            space_handle=space_handle,
        )

    def delete(self) -> Response[Tag]:
        return self.client.post(
            "tag/delete",
            Tag.DeleteRequest(id=self.id, file_id=self.file_id, block_id=self.block_id),
            expect=Tag,
        )

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[TagQueryResponse]:
        req = TagQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "tag/query",
            payload=req,
            expect=TagQueryResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )
        return res


class TagQueryResponse(Response):
    tags: List[Tag]


Tag.ListResponse.update_forward_refs()



from typing import Callable, Optional, Type

import pytest
from steamship_tests.utils.client import get_steamship_client

from steamship import Space, Steamship
from steamship.app.app import App
from steamship.app.lambda_handler import create_handler as _create_handler
from steamship.app.request import Invocation, InvocationContext, LoggingConfig, Request


@pytest.fixture()
def client() -> Steamship:
    """Returns a client rooted in a new space, then deletes that space afterwards.

    To use, simply import this file and then write a test which takes `client`
    as an argument.

    Example
    -------
    The client can be used by injecting a fixture as follows::

        @pytest.mark.usefixtures("client")
        def test_something(client):
          pass
    """
    steamship = get_steamship_client()
    space = Space.create(client=steamship).data
    new_client = steamship.for_space(space_id=space.id)
    yield new_client
    space.delete()


@pytest.fixture()
def app_handler(request) -> Callable[[str, str, Optional[dict]], dict]:
    """
    Returns a client rooted in a new space, then deletes that space afterwards.

    To use, simply import this file and then write a test which takes `app_handler`
    as an argument and parameterize it via PyTest.

    Example
    --------

        import pytest # doctest: +SKIP
        from steamship_tests.utils.fixtures import app_handler  # noqa: F401
        from assets.apps.demo_app import TestApp

        @pytest.mark.parametrize("app_handler", [TestApp], indirect=True)
            def _test_something(app_handler):
                response_dict = app_handler("POST", "/hello", dict())

    The app will be run its own space that gets cleaned up afterwards, and
    the test can be written from the perspective of an external caller of the
    app.
    """
    app: Type[App] = request.param
    steamship = get_steamship_client()
    space = Space.create(client=steamship).data
    new_client = steamship.for_space(space_id=space.id)

    def handle(verb: str, app_path: str, arguments: Optional[dict] = None) -> dict:
        _handler = _create_handler(app)
        invocation = Invocation(http_verb=verb, app_path=app_path, arguments=arguments or {})
        logging_config = LoggingConfig(logging_host="none", logging_port="none")
        request = Request(
            client_config=new_client.config,
            invocation=invocation,
            logging_config=logging_config,
            invocation_context=InvocationContext(),
        )
        event = request.dict(by_alias=True)
        return _handler(event)

    yield handle
    space.delete()



from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel

from steamship.base import Client, Request, Response, metadata_to_str
from steamship.base.configuration import CamelModel
from steamship.data.search import Hit


class EmbedAndSearchRequest(Request):
    query: str
    docs: List[str]
    plugin_instance: str
    k: int = 1


# TODO: These types are not generics like the Swift QueryResult/QueryResults.
class QueryResult(CamelModel):
    value: Optional[Hit] = None
    score: Optional[float] = None
    index: Optional[int] = None
    id: Optional[str] = None


class QueryResults(Request):
    items: List[QueryResult] = None


class EmbeddedItem(CamelModel):
    id: str = None
    index_id: str = None
    file_id: str = None
    block_id: str = None
    tag_id: str = None
    value: str = None
    external_id: str = None
    external_type: str = None
    metadata: Any = None
    embedding: List[float] = None

    def clone_for_insert(self) -> EmbeddedItem:
        """Produces a clone with a string representation of the metadata"""
        ret = EmbeddedItem(
            id=self.id,
            index_id=self.index_id,
            file_id=self.file_id,
            block_id=self.block_id,
            tag_id=self.tag_id,
            value=self.value,
            external_id=self.external_id,
            external_type=self.external_type,
            metadata=self.metadata,
            embedding=self.embedding,
        )
        if isinstance(ret.metadata, dict) or isinstance(ret.metadata, list):
            ret.metadata = json.dumps(ret.metadata)
        return ret


class IndexCreateRequest(Request):
    handle: str = None
    name: str = None
    plugin_instance: str = None
    upsert: bool = True
    external_id: str = None
    external_type: str = None
    metadata: Any = None


class IndexInsertRequest(Request):
    index_id: str
    items: List[EmbeddedItem] = None
    value: str = None
    file_id: str = None
    block_type: str = None
    external_id: str = None
    external_type: str = None
    metadata: Any = None
    reindex: bool = True


class IndexItemId(CamelModel):
    index_id: str = None
    id: str = None


class IndexInsertResponse(Response):
    item_ids: List[IndexItemId] = None


class IndexEmbedRequest(Request):
    id: str


class IndexEmbedResponse(Response):
    id: Optional[str] = None


class IndexSearchRequest(Request):
    id: str
    query: str = None
    queries: List[str] = None
    k: int = 1
    include_metadata: bool = False


class IndexSnapshotRequest(Request):
    index_id: str
    window_size: int = None  # Used for unit testing


class IndexSnapshotResponse(Response):
    id: Optional[str] = None
    snapshot_id: str


class ListSnapshotsRequest(Request):
    id: str = None


class ListSnapshotsResponse(Response):
    snapshots: List[IndexSnapshotResponse]


class ListItemsRequest(Request):
    id: str = None
    file_id: str = None
    block_id: str = None
    span_id: str = None


class ListItemsResponse(Response):
    items: List[EmbeddedItem]


class DeleteSnapshotsRequest(Request):
    snapshot_id: str = None


class DeleteSnapshotsResponse(Response):
    snapshot_id: str = None


class DeleteEmbeddingIndexRequest(Request):
    id: str


class EmbeddingIndex(CamelModel):
    """A persistent, read-optimized index over embeddings."""

    client: Client = None
    id: str = None
    handle: str = None
    name: str = None
    plugin: str = None
    external_id: str = None
    external_type: str = None
    metadata: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        if "embeddingIndex" in obj:
            obj = obj["embeddingIndex"]
        elif "index" in obj:
            obj = obj["index"]
        return super().parse_obj(obj)

    def insert_file(
        self,
        file_id: str,
        block_type: str = None,
        external_id: str = None,
        external_type: str = None,
        metadata: Union[int, float, bool, str, List, Dict] = None,
        reindex: bool = True,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[IndexInsertResponse]:
        if isinstance(metadata, dict) or isinstance(metadata, list):
            metadata = json.dumps(metadata)

        req = IndexInsertRequest(
            index_id=self.id,
            file_id=file_id,
            blockType=block_type,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata,
            reindex=reindex,
        )
        return self.client.post(
            "embedding-index/item/create",
            req,
            expect=IndexInsertResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def insert_many(
        self,
        items: List[Union[EmbeddedItem, str]],
        reindex: bool = True,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[IndexInsertResponse]:
        new_items = []
        for item in items:
            if isinstance(item, str):
                new_items.append(EmbeddedItem(value=item))
            else:
                new_items.append(item)

        req = IndexInsertRequest(
            index_id=self.id,
            items=[item.clone_for_insert() for item in new_items],
            reindex=reindex,
        )
        return self.client.post(
            "embedding-index/item/create",
            req,
            expect=IndexInsertResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def insert(
        self,
        value: str,
        external_id: str = None,
        external_type: str = None,
        metadata: Union[int, float, bool, str, List, Dict] = None,
        reindex: bool = True,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[IndexInsertResponse]:

        req = IndexInsertRequest(
            index_id=self.id,
            value=value,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata_to_str(metadata),
            reindex=reindex,
        )
        return self.client.post(
            "embedding-index/item/create",
            req,
            expect=IndexInsertResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def embed(
        self, space_id: str = None, space_handle: str = None, space: Any = None
    ) -> Response[IndexEmbedResponse]:
        req = IndexEmbedRequest(id=self.id)
        return self.client.post(
            "embedding-index/embed",
            req,
            expect=IndexEmbedResponse,
            asynchronous=True,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def create_snapshot(
        self, space_id: str = None, space_handle: str = None, space: Any = None
    ) -> Response[IndexSnapshotResponse]:
        req = IndexSnapshotRequest(index_id=self.id)
        return self.client.post(
            "embedding-index/snapshot/create",
            req,
            expect=IndexSnapshotResponse,
            asynchronous=True,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    # TODO (enias): Can these be generic list operations for all file types?
    def list_snapshots(
        self, space_id: str = None, space_handle: str = None, space: Any = None
    ) -> Response[ListSnapshotsResponse]:
        req = ListSnapshotsRequest(id=self.id)
        return self.client.post(
            "embedding-index/snapshot/list",
            req,
            expect=ListSnapshotsResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def list_items(
        self,
        file_id: str = None,
        block_id: str = None,
        span_id: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[ListItemsResponse]:
        req = ListItemsRequest(id=self.id, file_id=file_id, block_id=block_id, spanId=span_id)
        return self.client.post(
            "embedding-index/item/list",
            req,
            expect=ListItemsResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def delete_snapshot(
        self,
        snapshot_id: str,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[DeleteSnapshotsResponse]:
        req = DeleteSnapshotsRequest(snapshotId=snapshot_id)
        return self.client.post(
            "embedding-index/snapshot/delete",
            req,
            expect=DeleteSnapshotsResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def delete(
        self, space_id: str = None, space_handle: str = None, space: Any = None
    ) -> Response[EmbeddingIndex]:
        return self.client.post(
            "embedding-index/delete",
            DeleteEmbeddingIndexRequest(id=self.id),
            expect=EmbeddingIndex,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def search(
        self,
        query: Union[str, List[str]],
        k: int = 1,
        include_metadata: bool = False,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[QueryResults]:
        if isinstance(query, list):
            req = IndexSearchRequest(
                id=self.id, queries=query, k=k, include_metadata=include_metadata
            )
        else:
            req = IndexSearchRequest(
                id=self.id, query=query, k=k, include_metadata=include_metadata
            )
        ret = self.client.post(
            "embedding-index/search",
            req,
            expect=QueryResults,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

        return ret

    @staticmethod
    def create(
        client: Client,
        handle: str = None,
        name: str = None,
        plugin_instance: str = None,
        upsert: bool = True,
        external_id: str = None,
        external_type: str = None,
        metadata: Any = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[EmbeddingIndex]:
        req = IndexCreateRequest(
            handle=handle,
            name=name,
            plugin_instance=plugin_instance,
            upsert=upsert,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata,
        )
        return client.post(
            "embedding-index/create",
            req,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
            expect=EmbeddingIndex,
        )



from __future__ import annotations

from typing import Any, Dict, Optional, Type, Union

from pydantic import BaseModel

from steamship.base import Client, Request, Response
from steamship.base.configuration import CamelModel
from steamship.data.plugin import (
    HostingCpu,
    HostingEnvironment,
    HostingMemory,
    HostingTimeout,
    HostingType,
)
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.raw_data_plugin_output import RawDataPluginOutput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput

from .block import Block
from .file import File
from .operations.tagger import TagRequest, TagResponse


class GetPluginInstanceRequest(Request):
    id: Optional[str] = None
    handle: Optional[str] = None


class CreatePluginInstanceRequest(Request):
    id: str = None
    plugin_id: str = None
    plugin_handle: str = None
    plugin_version_id: str = None
    plugin_version_handle: str = None
    handle: str = None
    upsert: bool = None
    config: Dict[str, Any] = None


class DeletePluginInstanceRequest(Request):
    id: str


class PluginInstance(CamelModel):
    client: Client = None
    id: str = None
    handle: str = None
    plugin_id: str = None
    plugin_version_id: str = None
    space_id: Optional[str] = None
    user_id: str = None
    config: Dict[str, Any] = None
    hosting_type: Optional[HostingType] = None
    hosting_cpu: Optional[HostingCpu] = None
    hosting_memory: Optional[HostingMemory] = None
    hosting_timeout: Optional[HostingTimeout] = None
    hosting_environment: Optional[HostingEnvironment] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["pluginInstance"] if "pluginInstance" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        space_id: str = None,
        plugin_id: str = None,
        plugin_handle: str = None,
        plugin_version_id: str = None,
        plugin_version_handle: str = None,
        handle: str = None,
        upsert: bool = True,
        config: Dict[str, Any] = None,
    ) -> Response[PluginInstance]:
        # TODO (enias): Document
        """Create a plugin instance

        When handle is empty the engine will automatically assign one
        upsert controls whether we want to re-use an existing plugin instance or not."""
        req = CreatePluginInstanceRequest(
            handle=handle,
            plugin_id=plugin_id,
            plugin_handle=plugin_handle,
            plugin_version_id=plugin_version_id,
            plugin_version_handle=plugin_version_handle,
            upsert=upsert,
            config=config,
        )

        return client.post(
            "plugin/instance/create",
            payload=req,
            expect=PluginInstance,
            space_id=space_id,
        )

    @staticmethod
    def get(client: Client, handle: str):
        return client.post(
            "plugin/instance/get", GetPluginInstanceRequest(handle=handle), expect=PluginInstance
        )

    def tag(
        self,
        doc: Union[str, File],
    ) -> Response[TagResponse]:
        req = TagRequest(
            type="inline",
            file=File.CreateRequest(blocks=[Block.CreateRequest(text=doc)])
            if isinstance(doc, str)
            else doc,
            plugin_instance=self.handle,
        )
        return self.client.post(
            "plugin/instance/tag",
            req,
            expect=TagResponse,
        )

    def delete(self) -> PluginInstance:
        req = DeletePluginInstanceRequest(id=self.id)
        return self.client.post("plugin/instance/delete", payload=req, expect=PluginInstance)

    def export(self, plugin_input: ExportPluginInput) -> Response[RawDataPluginOutput]:
        plugin_input.plugin_instance = self.handle
        return self.client.post(
            "plugin/instance/export", payload=plugin_input, expect=RawDataPluginOutput
        )

    def train(self, training_request: TrainingParameterPluginInput) -> Response[TrainPluginOutput]:
        return self.client.post(
            "plugin/instance/train",
            space_id=self.space_id,
            payload=training_request,
            expect=TrainPluginOutput,
        )

    def get_training_parameters(
        self, training_request: TrainingParameterPluginInput
    ) -> Response[TrainingParameterPluginOutput]:
        return self.client.post(
            "plugin/instance/getTrainingParameters",
            payload=training_request,
            expect=TrainingParameterPluginOutput,
        )


class ListPrivatePluginInstancesRequest(Request):
    pass



from typing import Optional


class DocTag:
    """This class provides a set of Markdown-style constants to use with Tag(kind="doc")"""

    doc = "doc"
    page = "page"  # E.g. in a PDF
    region = "region"  # E.g., abstract catchall region in a document
    header = "header"
    h1 = "h1"
    h2 = "h2"
    h3 = "h3"
    h4 = "h4"
    h5 = "h5"
    line = "line"
    title = "title"
    subtitle = "subtitle"
    footer = "footer"
    paragraph = "paragraph"
    list = "list"
    list_item = "listitem"
    link = "link"
    caption = "caption"
    image = "image"
    blockquote = "blockquote"
    blockcode = "blockcode"
    unk = "unk"
    sentence = "sentence"
    token = "token"  # noqa: S105
    span = "span"
    div = "div"
    pre = "pre"
    strong = "strong"
    emph = "emph"
    underline = "underline"
    teletype = "teletype"

    @staticmethod
    def from_html_tag(tagname: Optional[str]) -> Optional[str]:  # noqa: C901
        if tagname is None:
            return None

        name = tagname.lower().strip()

        if name == "p":
            return DocTag.paragraph
        elif name == "h1":
            return DocTag.h1
        elif name == "h2":
            return DocTag.h2
        elif name == "h3":
            return DocTag.h3
        elif name == "h4":
            return DocTag.h4
        elif name == "h5":
            return DocTag.h5
        elif name == "ul":
            return DocTag.list
        elif name == "li":
            return DocTag.list_item
        elif name == "a":
            return DocTag.link
        elif name == "div":
            return DocTag.div
        elif name == "img":
            return DocTag.image
        elif name == "span":
            return DocTag.span
        elif name == "pre":
            return DocTag.pre
        elif name == "code":
            return DocTag.blockcode
        elif name == "blockquote":
            return DocTag.blockquote
        elif name == "strong":
            return DocTag.strong
        elif name == "b":
            return DocTag.strong
        elif name == "emph":
            return DocTag.emph
        elif name == "i":
            return DocTag.emph
        elif name == "u":
            return DocTag.underline
        elif name == "tt":
            return DocTag.teletype

        return None



from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from .base import Configuration, MimeTypes, SteamshipError
from .data import (
    App,
    AppInstance,
    AppVersion,
    Block,
    DocTag,
    EmbeddingIndex,
    File,
    PluginInstance,
    PluginVersion,
    Space,
    Tag,
    TagKind,
    TextTag,
)

from .client import Steamship  # isort:skip

__all__ = [
    "Steamship",
    "Configuration",
    "SteamshipError",
    "MimeTypes",
    "App",
    "AppInstance",
    "AppVersion",
    "File",
    "Block",
    "Tag",
    "Space",
    "PluginInstance",
    "PluginVersion",
    "TagKind",
    "TextTag",
    "DocTag",
    "EmbeddingIndex",
]



# Plugin
#
# This file contains the abstractions for managing Steamship plugins.
# To see how to implement a Steamship Plugin, see service.py in the same folder.
#
#

from __future__ import annotations

import json
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel

from steamship.base.client import Client
from steamship.base.configuration import CamelModel
from steamship.base.request import Request
from steamship.base.response import Response


class HostingType(str, Enum):
    """The type of hosting provider to deploy to."""

    LAMBDA = "lambda"
    ECS = "ecs"


class HostingEnvironment(str, Enum):
    """The software environment required for deployment."""

    PYTHON38 = "python38"
    STEAMSHIP_PYTORCH_CPU = "inferenceCpu"


class HostingMemory(str, Enum):
    """The amount of memory required for deployment.

    This is mapped to a value dependent on the HostingType it is combined with.
    """

    MIN = "min"
    XXS = "xxs"
    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"
    XL = "xl"
    XXL = "xxl"
    MAX = "max"


class HostingCpu(str, Enum):
    """The amount of CPU required for deployment.

    This is mapped to a value dependent on the HostingType it is combined with.
    """

    MIN = "min"
    XXS = "xxs"
    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"
    XL = "xl"
    XXL = "xxl"
    MAX = "max"


class HostingTimeout(str, Enum):
    """The request timeout required for deployment.

    This is mapped to a value dependent on the HostingType it is combined with.
    """

    MIN = "min"
    XXS = "xxs"
    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"
    XL = "xl"
    XXL = "xxl"
    MAX = "max"


class CreatePluginRequest(Request):
    training_platform: Optional[HostingType] = None
    id: str = None
    type: str = None
    transport: str = None
    is_public: bool = None
    handle: str = None
    description: str = None
    metadata: str = None
    upsert: bool = None


class DeletePluginRequest(Request):
    id: str


class ListPublicPluginsRequest(Request):
    type: Optional[str] = None


class ListPrivatePluginsRequest(Request):
    type: str


class ListPluginsResponse(Response):
    plugins: List[Plugin]


class GetPluginRequest(Request):
    type: str = None
    id: str = None
    handle: str = None


class PluginType(str, Enum):
    parser = "parser"
    classifier = "classifier"
    tagger = "tagger"
    embedder = "embedder"


class PluginAdapterType(str, Enum):
    steamship_docker = "steamshipDocker"
    steamship_sagemaker = "steamshipSagemaker"
    huggingface = "huggingface"
    openai = "openai"


class PluginTargetType(str, Enum):
    file = "file"
    space = "space"


class LimitUnit(str, Enum):
    words = "words"
    characters = "characters"
    bytes = "bytes"


class Plugin(CamelModel):
    # TODO (enias): Document, Plugins are just a name placeholder with some settings
    client: Client = None
    id: str = None
    type: str = None
    transport: str = None
    is_public: bool = None
    training_platform: Optional[HostingType] = None
    handle: str = None
    description: str = None
    metadata: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["plugin"] if "plugin" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        description: str,
        type_: str,
        transport: str,
        is_public: bool,
        handle: str = None,
        training_platform: Optional[HostingType] = None,
        metadata: Union[str, Dict, List] = None,
        upsert: bool = False,
        space_id: str = None,
        space_handle: str = None,
    ) -> Response[Plugin]:
        if isinstance(metadata, dict) or isinstance(metadata, list):
            metadata = json.dumps(metadata)

        req = CreatePluginRequest(
            training_platform=training_platform,
            type=type_,
            transport=transport,
            is_public=is_public,
            handle=handle,
            description=description,
            metadata=metadata,
            upsert=upsert,
        )
        return client.post(
            "plugin/create",
            req,
            expect=Plugin,
            space_id=space_id,
            space_handle=space_handle,
        )

    @staticmethod
    def list(
        client: Client, t: str = None, space_id: str = None, space_handle: str = None
    ) -> Response[ListPluginsResponse]:
        return client.post(
            "plugin/list",
            ListPublicPluginsRequest(type=t),
            expect=ListPluginsResponse,
            space_id=space_id,
            space_handle=space_handle,
        )

    @staticmethod
    def get(client: Client, handle: str):
        return client.post("plugin/get", GetPluginRequest(handle=handle), expect=Plugin)

    def delete(self) -> Response[Plugin]:
        return self.client.post(
            "plugin/delete",
            DeletePluginRequest(id=self.id),
            expect=Plugin,
        )


ListPluginsResponse.update_forward_refs()



# noinspection PyPackageRequirements
import uuid

import pytest
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import PluginInstance, Space
from steamship.data.plugin import Plugin, PluginAdapterType, PluginType


def test_plugin_create():
    steamship = get_steamship_client()

    my_plugins = Plugin.list(steamship).data
    orig_count = len(my_plugins.plugins)

    plugin_args = {
        "client": steamship,
        "description": "This is just for test",
        "type_": PluginType.embedder,
        "transport": PluginAdapterType.steamship_docker,
        "is_public": True,
    }

    required_fields = {"description", "type_"}
    for required_field in required_fields:
        del plugin_args[required_field]
        with pytest.raises(TypeError):
            Plugin.create(**plugin_args)

    my_plugins = Plugin.list(steamship).data
    assert len(my_plugins.plugins) == orig_count

    plugin = Plugin.create(
        client=steamship,
        description="This is just for test",
        type_=PluginType.tagger,
        transport=PluginAdapterType.steamship_docker,
        is_public=False,
    ).data
    my_plugins = Plugin.list(steamship).data
    assert len(my_plugins.plugins) == orig_count + 1

    # No upsert doesn't work
    plugin_x = Plugin.create(
        client=steamship,
        handle=plugin.handle,
        description="This is just for test",
        type_=PluginType.tagger,
        transport=PluginAdapterType.steamship_docker,
        is_public=False,
        upsert=False,
    )
    assert plugin_x.error is not None

    # Upsert does work
    plugin2 = Plugin.create(
        client=steamship,
        handle=plugin.handle,
        description="This is just for test 2",
        type_=PluginType.tagger,
        transport=PluginAdapterType.steamship_docker,
        is_public=False,
        upsert=True,
    ).data

    assert plugin2.id == plugin.id

    my_plugins = Plugin.list(steamship).data
    assert len(my_plugins.plugins) == orig_count + 1

    assert plugin2.id in [plugin.id for plugin in my_plugins.plugins]
    assert plugin2.description in [plugin.description for plugin in my_plugins.plugins]

    # assert(my_plugins.plugins[0].description != plugin.description)

    plugin.delete()

    my_plugins = Plugin.list(steamship).data
    assert len(my_plugins.plugins) == orig_count


def test_plugin_public():
    steamship = get_steamship_client()

    resp = Plugin.list(steamship).data
    assert resp.plugins is not None
    plugins = resp.plugins

    assert len(plugins) > 0

    # Make sure they can't be deleted.
    res = plugins[0].delete()
    assert res.error is not None


def test_deploy_in_space():
    client = get_steamship_client()
    space = Space.create(client, handle="test-non-default-space").data
    instance = PluginInstance.create(client, plugin_handle="test-tagger", space_id=space.id).data
    assert instance.space_id == space.id


def test_plugin_instance_get():
    steamship = get_steamship_client()
    handle = f"test_tagger_test_handle{uuid.uuid4()}"
    instance = PluginInstance.create(steamship, plugin_handle="test-tagger", handle=handle).data
    assert instance.id is not None

    other_instance = PluginInstance.get(steamship, handle=handle).data

    assert instance.id == other_instance.id


def test_plugin_instance_quick_create():
    steamship = get_steamship_client()

    handle = f"test_tagger_test_handle{uuid.uuid4()}"
    handle2 = f"test_tagger_test_handle{uuid.uuid4()}"

    p1 = steamship.use_plugin("test-tagger", handle)
    p2 = steamship.use_plugin("test-tagger", handle2)
    p3 = steamship.use_plugin("test-tagger", handle)
    p4 = steamship.use_plugin("test-tagger")

    assert p1.id == p3.id
    assert p1.id != p2.id
    assert p1.id != p4.id



from .app import App, get, post
from .lambda_handler import create_handler
from .request import Invocation, Request
from .response import Response

__all__ = ["App", "create_handler", "Invocation", "Request", "Response", "get", "post"]



from abc import ABC, abstractmethod

from steamship.app import Response, post
from steamship.data.operations.corpus_importer import CorpusImportRequest, CorpusImportResponse
from steamship.plugin.service import PluginRequest, PluginService


# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a Corpus Importer.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.corpus_importer instead
# of this file.
#
class CorpusImporter(PluginService[CorpusImportRequest, CorpusImportResponse], ABC):
    @abstractmethod
    def run(self, request: PluginRequest[CorpusImportRequest]) -> Response[CorpusImportResponse]:
        raise NotImplementedError()

    @post("import")
    def run_endpoint(self, **kwargs) -> Response[CorpusImportResponse]:
        """Exposes the Corpus Importer's `run` operation to the Steamship Engine via the expected HTTP path POST /import"""
        return self.run(PluginRequest[CorpusImportRequest].parse_obj(kwargs))



import contextlib
import io
import logging
import os
import subprocess  # noqa: S404
import tempfile
import time
import zipfile
from pathlib import Path
from typing import Any, Dict, Optional

from steamship_tests import SRC_PATH, TEST_ASSETS_PATH

from steamship import App, AppInstance, AppVersion, Steamship
from steamship.data.plugin import HostingType, Plugin
from steamship.data.plugin_instance import PluginInstance
from steamship.data.plugin_version import PluginVersion
from steamship.data.user import User


def install_package(package: str, into_folder: str):
    logging.info(f"Installing {package} into: {into_folder}")
    subprocess.run(  # noqa: S607,S603
        ["pip", "install", "--target", into_folder, package], stdout=subprocess.PIPE
    )


def zip_deployable(file_path: Path) -> bytes:
    """Prepare and zip a Steamship plugin."""

    package_paths = [
        SRC_PATH / "steamship",
        SRC_PATH
        / ".."
        / "steamship_tests",  # This is included to test plugin development using inheritance
    ]

    zip_buffer = io.BytesIO()
    with zipfile.ZipFile(
        file=zip_buffer, mode="a", compression=zipfile.ZIP_DEFLATED, allowZip64=False
    ) as zip_file:
        zip_file.write(file_path, "api.py")

        # Copy in the package paths from source
        for package_path in package_paths:
            for root, _, files in os.walk(package_path):
                for file in files:
                    pypi_file = Path(root) / file
                    zip_file.write(pypi_file, pypi_file.relative_to(package_path.parent))

        # Copy in package paths from pip
        with tempfile.TemporaryDirectory() as package_dir:
            logging.info(f"Created tempdir for pip installs: {package_dir}")
            for package in [
                "setuptools_scm",
                "requests",
                "charset_normalizer",
                "certifi",
                "urllib3",
                "idna",
                "pydantic==1.9.0",
                "typing_extensions",
                "inflection==0.5.1",
                "fluent-logger==0.10.0",
            ]:
                install_package(package, into_folder=package_dir)
            # Now write that whole folder
            for root, _, files in os.walk(package_dir):
                for file in files:
                    pypi_file = Path(root) / file
                    zip_file.write(pypi_file, pypi_file.relative_to(package_dir))

        # Now we'll copy in the whole assets directory so that our test files can access things there..
        for root, _, files in os.walk(TEST_ASSETS_PATH):
            for file in files:
                the_file = Path(root) / file
                relative_path = the_file.relative_to(TEST_ASSETS_PATH.parent)
                zip_file.write(the_file, relative_path)

    # Leaving this in as a reminder: this is an easy way to generate the app zip for use in
    # updating engine unit steamship_tests.
    #
    # with open("demo_app.zip", 'wb') as f:
    #     f.write(zip_buffer.getvalue())
    #

    return zip_buffer.getvalue()


@contextlib.contextmanager
def deploy_plugin(
    client: Steamship,
    py_path: Path,
    plugin_type: str,
    training_platform: Optional[HostingType] = None,
    version_config_template: Dict[str, Any] = None,
    instance_config: Dict[str, Any] = None,
    space_id: Optional[str] = None,
):
    plugin = Plugin.create(
        client,
        training_platform=training_platform,
        type_=plugin_type,
        transport="jsonOverHttp",
        description="A Plugin (python client steamship_tests)",
        is_public=False,
    )
    assert plugin.error is None
    assert plugin.data is not None
    plugin = plugin.data

    zip_bytes = zip_deployable(py_path)
    version = PluginVersion.create(
        client,
        "test-version",
        plugin_id=plugin.id,
        filebytes=zip_bytes,
        config_template=version_config_template,
    )
    # TODO: This is due to having to wait for the lambda to finish deploying.
    # TODO: We should update the task system to allow its .wait() to depend on this.
    version = _wait_for_version(version)

    instance = PluginInstance.create(
        client,
        space_id=space_id,
        plugin_id=plugin.id,
        plugin_version_id=version.id,
        config=instance_config,
    )
    instance = _wait_for_instance(instance)

    assert instance.plugin_id == plugin.id
    assert instance.plugin_version_id == version.id

    _check_user(client, instance)

    yield plugin, version, instance

    _delete_deployable(instance, version, plugin)


@contextlib.contextmanager
def deploy_app(
    client: Steamship,
    py_path: Path,
    version_config_template: Dict[str, Any] = None,
    instance_config: Dict[str, Any] = None,
    space_id: str = None,
):
    app = App.create(client)
    assert app.error is None
    assert app.data is not None
    app = app.data

    zip_bytes = zip_deployable(py_path)
    version = AppVersion.create(
        client,
        app_id=app.id,
        filebytes=zip_bytes,
        config_template=version_config_template,
    )

    version = _wait_for_version(version)
    instance = AppInstance.create(
        client,
        app_id=app.id,
        app_version_id=version.id,
        config=instance_config,
        space_id=space_id,
    )
    instance = _wait_for_instance(instance)

    assert instance.app_id == app.id
    assert instance.app_version_id == version.id

    _check_user(client, instance)

    yield app, version, instance

    _delete_deployable(instance, version, app)


def _check_user(client, instance):
    user = User.current(client).data
    assert instance.user_id == user.id


def _delete_deployable(instance, version, deployable):
    res = instance.delete()
    assert res.error is None
    res = version.delete()
    assert res.error is None
    res = deployable.delete()
    assert res.error is None


def _wait_for_instance(instance):
    instance.wait()
    assert instance.error is None
    assert instance.data is not None
    return instance.data


def _wait_for_version(version):
    time.sleep(15)
    version.wait()
    assert version.error is None
    assert version.data is not None
    return version.data



from .client import Client
from .configuration import Configuration
from .error import SteamshipError
from .metadata import metadata_to_str, str_to_metadata
from .mime_types import MimeTypes
from .request import Request
from .response import Response
from .tasks import Task, TaskComment, TaskState

__all__ = [
    "Client",
    "Configuration",
    "SteamshipError",
    "metadata_to_str",
    "str_to_metadata",
    "MimeTypes",
    "Request",
    "Response",
    "Task",
    "TaskComment",
    "TaskState",
]



from .app import App
from .app_instance import AppInstance
from .app_version import AppVersion
from .block import Block
from .embeddings import EmbeddingIndex
from .file import File
from .plugin import Plugin
from .plugin_instance import PluginInstance
from .plugin_version import PluginVersion
from .space import Space
from .tags import DocTag, Tag, TagKind, TextTag

__all__ = [
    "App",
    "AppInstance",
    "AppVersion",
    "Block",
    "EmbeddingIndex",
    "File",
    "Plugin",
    "PluginInstance",
    "PluginVersion",
    "Space",
    "DocTag",
    "Tag",
    "TagKind",
    "TextTag",
]



from __future__ import annotations

import io
import logging
from enum import Enum
from typing import Any, List, Optional, Type, Union

from pydantic import BaseModel

from steamship.base import Client, Request, Response
from steamship.base.binary_utils import flexi_create
from steamship.base.configuration import CamelModel
from steamship.base.request import IdentifierRequest
from steamship.data.block import Block
from steamship.data.embeddings import EmbeddingIndex
from steamship.data.tags import Tag


class FileUploadType(str, Enum):
    FILE = "file"  # The CreateRequest contains a file upload that should be used
    VALUE = "value"  # The Create Request contains a `text` field that should be used
    FILE_IMPORTER = (
        "fileImporter"  # The CreateRequest contains a fileImporter handle that should be used
    )
    BLOCKS = "blocks"  # The CreateRequest contains blocks and tags that should be read in directly


_logger = logging.getLogger(__name__)


class FileClearResponse(Response):
    id: str


class FileQueryRequest(Request):
    tag_filter_query: str


class File(CamelModel):
    """A file."""

    client: Client = None
    id: str = None
    handle: str = None
    mime_type: str = None
    space_id: str = None
    corpus_id: str = None
    blocks: List[Block] = []
    tags: List[Tag] = []
    filename: str = None

    class CreateRequest(Request):
        value: str = None
        data: str = None
        id: str = None
        url: str = None
        filename: str = None
        type: FileUploadType = None
        mime_type: str = None
        corpus_id: str = None
        blocks: Optional[List[Block.CreateRequest]] = []
        tags: Optional[List[Tag.CreateRequest]] = []
        plugin_instance: str = None

        class Config:
            use_enum_values = True

    class CreateResponse(Response):
        data_: Any = None
        mime_type: str = None

        def __init__(
            self,
            data: Any = None,
            string: str = None,
            _bytes: Union[bytes, io.BytesIO] = None,
            json: io.BytesIO = None,
            mime_type: str = None,
        ):
            super().__init__()
            data, mime_type, encoding = flexi_create(
                data=data, string=string, json=json, _bytes=_bytes, mime_type=mime_type
            )
            self.data_ = data
            self.mime_type = mime_type

        def to_dict(self) -> dict:
            return {"data": self.data_, "mime_type": self.mime_type}

    class ListRequest(Request):
        corpus_id: str = None

    class ListResponse(Response):
        files: List[File]

    class RawRequest(Request):
        id: str

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["file"] if "file" in obj else obj
        return super().parse_obj(obj)

    def delete(
        self, space_id: str = None, space_handle: str = None, space: Any = None
    ) -> Response[File]:
        return self.client.post(
            "file/delete",
            IdentifierRequest(id=self.id),
            expect=File,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def clear(
        self, space_id: str = None, space_handle: str = None, space: Any = None
    ) -> Response[FileClearResponse]:
        return self.client.post(
            "file/clear",
            IdentifierRequest(id=self.id),
            expect=FileClearResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    @staticmethod
    def get(
        client: Client,
        _id: str = None,
        handle: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[File]:  # TODO (Enias): Why is this a staticmethod?
        return client.post(
            "file/get",
            IdentifierRequest(id=_id, handle=handle),
            expect=File,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    @staticmethod
    def create(
        client: Client,
        filename: str = None,
        url: str = None,
        content: str = None,
        plugin_instance: str = None,
        mime_type: str = None,
        blocks: List[Block.CreateRequest] = None,
        tags: List[Tag.CreateRequest] = None,
        corpus_id: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[File]:

        if (
            filename is None
            and content is None
            and url is None
            and plugin_instance is None
            and blocks is None
        ):
            raise Exception("Either filename, content, url, or plugin Instance must be provided.")

        if blocks is not None:
            upload_type = FileUploadType.BLOCKS
        elif plugin_instance is not None:
            upload_type = FileUploadType.FILE_IMPORTER
        elif content is not None:
            # We're still going to use the file upload method for file uploads
            upload_type = FileUploadType.FILE
        elif filename is not None:
            with open(filename, "rb") as f:
                content = f.read()
            upload_type = FileUploadType.FILE
        else:
            raise Exception("Unable to determine upload type.")

        req = File.CreateRequest(
            type=upload_type,
            corpusId=corpus_id,
            url=url,
            mime_type=mime_type,
            plugin_instance=plugin_instance,
            blocks=blocks,
            tags=tags,
            filename=filename,
        )

        # Defaulting this here, as opposed to in the Engine, because it is processed by Vapor
        file_part_name = filename if filename else "unnamed"
        return client.post(
            "file/create",
            payload=req,
            file=(file_part_name, content, "multipart/form-data")
            if upload_type != FileUploadType.BLOCKS
            else None,
            expect=File,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    @staticmethod
    def list(
        client: Client,
        corpus_id: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ):
        req = File.ListRequest(corpusId=corpus_id)
        res = client.post(
            "file/list",
            payload=req,
            expect=File.ListResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )
        return res

    def refresh(self):
        return File.get(self.client, self.id)

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[FileQueryResponse]:

        req = FileQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "file/query",
            payload=req,
            expect=FileQueryResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )
        return res

    def raw(self, space_id: str = None, space_handle: str = None, space: Any = None):
        req = File.RawRequest(
            id=self.id,
        )
        # TODO (enias): Investigate why we do not need a expect here
        return self.client.post(
            "file/raw",
            payload=req,
            space_id=space_id or self.space_id,
            space_handle=space_handle,
            space=space,
            raw_response=True,
        )

    def blockify(self, plugin_instance: str = None):
        from steamship.data.operations.blockifier import BlockifyRequest
        from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

        req = BlockifyRequest(type="file", id=self.id, plugin_instance=plugin_instance)

        return self.client.post(
            "plugin/instance/blockify",
            payload=req,
            expect=BlockAndTagPluginOutput,
            asynchronous=True,
        )

    def tag(
        self,
        plugin_instance: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[Tag]:
        # TODO (enias): Fix Circular imports
        from steamship.data.operations.tagger import TagRequest, TagResponse
        from steamship.data.plugin import PluginTargetType

        req = TagRequest(type=PluginTargetType.file, id=self.id, plugin_instance=plugin_instance)
        return self.client.post(
            "plugin/instance/tag",
            payload=req,
            expect=TagResponse,
            asynchronous=True,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def index(
        self,
        plugin_instance: str = None,
        index_id: str = None,
        e_index: EmbeddingIndex = None,
        reindex: bool = True,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> EmbeddingIndex:
        # TODO: This should really be done all on the app, but for now we'll do it in the client
        # to facilitate demos.
        from steamship import EmbeddingIndex
        from steamship.data.embeddings import EmbeddedItem

        if index_id is None and e_index is not None:
            index_id = e_index.id

        if index_id is None and e_index is None:
            e_index = EmbeddingIndex.create(
                client=self.client,
                plugin_instance=plugin_instance,
                upsert=True,
                space_id=space_id,
                space_handle=space_handle,
                space=space,
            ).data
        elif e_index is None:
            e_index = EmbeddingIndex(client=self.client, id=index_id)

        # We have an index available to us now. Perform the query.
        blocks = self.refresh().data.blocks

        items = []
        for block in blocks:
            item = EmbeddedItem(value=block.text, external_id=block.id, external_type="block")
            items.append(item)

        insert_task = e_index.insert_many(
            items,
            reindex=reindex,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

        insert_task.wait()
        return e_index


class FileQueryResponse(Response):
    files: List[File]


File.ListResponse.update_forward_refs()
File.CreateRequest.update_forward_refs()



from __future__ import annotations

from typing import Any, List, Optional

from steamship.base import Client, Request, Response
from steamship.base.configuration import CamelModel
from steamship.base.request import IdentifierRequest
from steamship.data.tags.tag import Tag


class BlockQueryRequest(Request):
    tag_filter_query: str


class Block(CamelModel):
    client: Client = None
    id: str = None
    file_id: str = None
    text: str = None
    tags: Optional[List[Tag]] = []

    class CreateRequest(Request):
        id: str = None
        file_id: str = None
        text: str = None
        tags: Optional[List[Tag.CreateRequest]] = []
        upsert: bool = None

    class DeleteRequest(Request):
        id: str = None

    class ListRequest(Request):
        file_id: str = None

    class ListResponse(Response):
        blocks: List[Block] = []

    @staticmethod
    def get(
        client: Client,
        _id: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[Block]:
        return client.post(
            "block/get",
            IdentifierRequest(id=_id),
            expect=Block,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    @staticmethod
    def create(
        client: Client,
        file_id: str = None,
        text: str = None,
        tags: List[Tag.CreateRequest] = None,
        upsert: bool = None,
        space_id: str = None,
        space_handle: str = None,
    ) -> Response[Block]:
        req = Block.CreateRequest(file_id=file_id, text=text, tags=tags, upsert=upsert)
        return client.post(
            "block/create",
            req,
            expect=Block,
            space_id=space_id,
            space_handle=space_handle,
        )

    @staticmethod
    def list_public(
        client: Client,
        file_id: str = None,
        space_id: str = None,
        space_handle: str = None,
    ) -> Response[Block.ListResponse]:
        return client.post(
            "block/list",
            Block.ListRequest(file_id=file_id),
            expect=Block.ListResponse,
            space_id=space_id,
            space_handle=space_handle,
        )

    def delete(self) -> Response[Block]:
        return self.client.post(
            "block/delete",
            Block.DeleteRequest(id=self.id),
            expect=Tag,
        )

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response[BlockQueryResponse]:
        # TODO: Is this a static method?
        req = BlockQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "block/query",
            payload=req,
            expect=BlockQueryResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )
        return res


class BlockQueryResponse(Response):
    blocks: List[Block]


Block.ListResponse.update_forward_refs()
Block.update_forward_refs()



from __future__ import annotations

import json
import logging
import typing
from abc import ABC
from inspect import isclass
from typing import Any, Dict, Type, TypeVar, Union

import inflection
from pydantic import BaseModel, PrivateAttr
from requests import Session

from steamship.base.configuration import CamelModel, Configuration
from steamship.base.error import SteamshipError
from steamship.base.mime_types import MimeTypes
from steamship.base.request import Request
from steamship.base.response import Response, Task
from steamship.base.utils import to_camel
from steamship.utils.url import Verb, is_local

_logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Response)  # TODO (enias): Do we need this?


class Client(CamelModel, ABC):
    """Client base.py class.

    Separated primarily as a hack to prevent circular imports.
    """

    config: Configuration
    _session: Session = PrivateAttr()

    def __init__(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        web_base: str = None,
        space_id: str = None,
        space_handle: str = None,
        create_space: bool = False,
        profile: str = None,
        config_file: str = None,
        config: Configuration = None,
        **kwargs,
    ):
        if config is not None and not isinstance(config, Configuration):
            raise SteamshipError(
                f"config has to be of type {Configuration}, received {type(config)}."
            )
        config = config or Configuration(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            web_base=web_base,
            space_id=space_id,
            space_handle=space_handle,
            profile=profile,
            config_file=config_file,
        )
        self._session = Session()
        super().__init__(config=config)
        self.switch_space(space_id=space_id, space_handle=space_handle, create_space=create_space)

    def switch_space(
        self, space_id: str = None, space_handle: str = None, create_space: bool = False
    ):
        """Switches this client to the requested space, possibly creating it. If all arguments are None, the client
        actively switches into the default space.

        - API calls are performed manually to not result in circular imports.
        - Note that the default space is technically not necessary for API usage; it will be assumed by the Engine
          in the absense of a Space ID or Handle being manually specified in request headers.
        """
        return_id = None
        return_handle = None
        space = None

        if create_space:
            if space_id is not None:
                raise SteamshipError(
                    message="Can not initialize the Steamship client with create_space=True and also a space_id."
                )
            # If space_handle is none, one will be auto-generated for us.
            logging.info(f"Creating space with requested handle: '{return_handle}' .")

            # Zero out the space_handle on the config block in case we're being invoked from
            # `init` (otherwise we'll attempt to create the sapce IN that nonexistant space)
            old_space_handle = self.config.space_handle
            self.config.space_handle = None
            try:
                space = self.post(
                    "space/create",
                    {"handle": space_handle},
                    space_handle=None,  # self.config.space_handle doesn't exist yet!
                ).data
            except SteamshipError as e:
                self.config.space_handle = old_space_handle
                raise e

        else:
            if space_id is None and space_handle is None:
                # Switch to the default space
                space = self.post("space/get", {"handle": "default"}).data
            else:
                # Switch to the requested space
                space = self.post("space/get", {"handle": space_handle, "id": space_id}).data

        if space is None:
            raise SteamshipError(
                message="Was unable to switch to new space: server returned empty Space."
            )

        return_id = space.get("space", {}).get("id")
        return_handle = space.get("space", {}).get("handle")

        if return_id is None or return_handle is None:
            raise SteamshipError(
                message="Was unable to switch to new space: server returned empty ID and Handle."
            )

        # Finally, set the new space.
        self.config.space_id = return_id
        self.config.space_handle = return_handle
        logging.info(f"Switched to space '{return_handle}' (ID {return_id}).")

    def _url(
        self,
        is_app_call: bool = False,
        app_owner: str = None,
        operation: str = None,
    ):
        if not is_app_call:
            # Regular API call
            base = self.config.api_base
        else:
            # Do the app version
            if app_owner is None:
                return SteamshipError(
                    code="UserMissing",
                    message="Can not invoke an app endpoint without the app owner's user handle.",
                    suggestion="Provide the appOwner option, or initialize your app with a lookup.",
                )

            base = self.config.app_base
            if not is_local(base):
                # We want to prepend the user handle
                parts = base.split("//")
                base = f"{parts[0]}//{app_owner}.{'//'.join(parts[1:])}"

        # Clean leading and trailing "/"
        if base[len(base) - 1] == "/":
            base = base[:-1]
        if operation[0] == "/":
            operation = operation[1:]

        return f"{base}/{operation}"

    def _headers(
        self,
        space_id: str = None,
        space_handle: str = None,
        is_app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,
        as_background_task: bool = False,
    ):
        headers = {"Authorization": f"Bearer {self.config.api_key}"}

        if space_id is not None or space_handle is not None:
            sid = space_id
            shandle = space_handle
        else:
            sid = self.config.space_id
            shandle = self.config.space_handle

        if sid:
            headers["X-Space-Id"] = sid
        elif shandle:
            headers["X-Space-Handle"] = shandle

        if is_app_call:
            if app_owner:
                headers["X-App-Owner-Handle"] = app_owner
            if app_id:
                headers["X-App-Id"] = app_id
            if app_instance_id:
                headers["X-App-Instance-Id"] = app_instance_id

        if as_background_task:
            # Will result in the engine persisting the inbound HTTP request as a Task for deferred
            # execution. The client will receive task information back instead of the synchronous API response.
            # That task can be polled for eventual completion.
            headers["X-Task-Background"] = "true"

        return headers

    @staticmethod
    def _prepare_data(payload: Union[Request, dict]):
        if payload is None:
            data = {}
        elif isinstance(payload, dict):
            data = payload
        elif hasattr(payload, "to_dict"):
            data = payload.to_dict()
        elif isinstance(payload, BaseModel):
            data = payload.dict(by_alias=True)
        else:
            raise RuntimeError(f"Unable to parse payload of type {type(payload)}")

        return data

    @staticmethod
    def _response_data(resp, raw_response: bool = False):
        if resp is None:
            return None

        if raw_response:
            return resp.content

        if resp.headers:
            ct = None
            if "Content-Type" in resp.headers:
                ct = resp.headers["Content-Type"]
            if "content-type" in resp.headers:
                ct = resp.headers["content-type"]
            if ct is not None:
                ct = ct.split(";")[0]  # application/json; charset=utf-8
                if ct in [MimeTypes.TXT, MimeTypes.MKD, MimeTypes.HTML]:
                    return resp.text
                elif ct == MimeTypes.JSON:
                    return resp.json()
                else:
                    return resp.content

    @staticmethod
    def _prepare_multipart_data(data, file):
        # Note: requests seems to have a bug passing boolean (and maybe numeric?)
        # values in the midst of multipart form data. You need to manually convert
        # it to a string; otherwise it will pass as False or True (with the capital),
        # which is not standard notation outside of Python.
        for key in data:
            if data[key] is False:
                data[key] = "false"
            elif data[key] is True:
                data[key] = "true"

        result = {}
        for key, val in data.items():
            if val:
                if isinstance(val, Dict):
                    result[key] = (None, json.dumps(val), "application/json")
                else:
                    result[key] = (None, str(val))
        result["file"] = file
        return result

    def _add_client_to_response(self, expect: Type, response_data: Any):
        if isinstance(response_data, dict):
            self._add_client_to_object(expect, response_data)
        elif isinstance(response_data, list):
            for el in response_data:
                typing_parameters = typing.get_args(expect)
                self._add_client_to_response(
                    typing_parameters[0] if typing_parameters else None, el
                )

        return response_data

    def _add_client_to_object(self, expect, response_data):
        if expect and isclass(expect):
            if len(response_data.keys()) == 1 and list(response_data.keys())[0] in (
                to_camel(expect.__name__),
                "index",
            ):
                # TODO (enias): Hack since the engine responds with incosistent formats e.g. {"plugin" : {plugin_fields}}
                for _, v in response_data.items():
                    self._add_client_to_response(expect, v)
            elif issubclass(expect, BaseModel):
                response_data["client"] = self
                try:
                    key_to_type = typing.get_type_hints(expect)
                    for k, v in response_data.items():
                        self._add_client_to_response(key_to_type.get(inflection.underscore(k)), v)
                except NameError:
                    # typing.get_type_hints fails for Space
                    pass

    def call(  # noqa: C901
        self,
        verb: str,
        operation: str,
        payload: Union[Request, dict] = None,
        file: Any = None,
        expect: Type[T] = None,
        asynchronous: bool = False,
        debug: bool = False,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
        raw_response: bool = False,
        is_app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,  # TODO (Enias): Where is the app_version_id ?
        as_background_task: bool = False,
    ) -> Union[Any, Response[T]]:
        """Post to the Steamship API.

        All responses have the format::

        .. code-block:: json

        {
            "data": "<actual response>",
            "error": {"reason": "<message>"}
        } # noqa: RST203

        For the Python client we return the contents of the `data` field if present, and we raise an exception
        if the `error` field is filled in.
        """
        if space is not None:
            space_id = getattr(space, "id", None) if space_id is None else space_id
            space_handle = getattr(space, "handle", None) if space_handle is None else space_handle

        url = self._url(
            is_app_call=is_app_call,
            app_owner=app_owner,
            operation=operation,
        )

        headers = self._headers(
            space_id=space_id,
            space_handle=space_handle,
            is_app_call=is_app_call,
            app_owner=app_owner,
            app_id=app_id,
            app_instance_id=app_instance_id,
            as_background_task=as_background_task,
        )

        data = self._prepare_data(payload=payload)

        logging.info(f"Steamship Client making {verb} to {url}")
        if verb == Verb.POST:
            if file is not None:
                files = self._prepare_multipart_data(data, file)
                resp = self._session.post(url, files=files, headers=headers)
            else:
                resp = self._session.post(url, json=data, headers=headers)
        elif verb == Verb.GET:
            resp = self._session.get(url, params=data, headers=headers)
        else:
            raise Exception(f"Unsupported verb: {verb}")

        logging.info(f"Steamship Client received HTTP {resp.status_code} from {verb} to {url}")

        if debug is True:
            logging.debug(f"Got response {resp}")

        response_data = self._response_data(resp, raw_response=raw_response)

        logging.debug(f"Response JSON {response_data}")

        task = None
        error = None

        if isinstance(response_data, dict):
            if "status" in response_data:
                task = Task.parse_obj({**response_data["status"], "client": self})
                # if task_resp is not None and task_resp.taskId is not None:
                #     task = Task(client=self)
                #     task.update(task_resp)
                if "state" in response_data["status"]:
                    if response_data["status"]["state"] == "failed":
                        error = SteamshipError.from_dict(response_data["status"])
                        logging.error(f"Client received error from server: {error}")

            if "data" in response_data:
                if expect is not None:
                    if hasattr(expect, "from_dict"):
                        data = expect.from_dict(response_data["data"], client=self)
                    # elif get_origin(expect) and issubclass(get_origin(expect), List):
                    #     if issubclass(expect.__args__[0], BaseModel):
                    #         parse_obj_as(expect, self._add_client_to_response( response_data["data"]))
                    elif issubclass(expect, BaseModel):
                        data = expect.parse_obj(
                            self._add_client_to_response(expect, response_data["data"])
                        )
                    else:
                        raise RuntimeError(f"obj of type {expect} does not have a from_dict method")
                else:
                    data = response_data["data"]
                    expect = type(data)
            else:
                data = response_data

        else:
            data = response_data
            expect = type(response_data)

        ret = Response(expect=expect, task=task, data_=data, error=error, client=self)
        if ret.task is None and ret.data is None and ret.error is None:
            raise Exception("No data, task status, or error found in response")

        return ret

    def post(
        self,
        operation: str,
        payload: Union[Request, dict, BaseModel] = None,
        file: Any = None,
        expect: Any = None,
        asynchronous: bool = False,
        debug: bool = False,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
        raw_response: bool = False,
        app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,
        as_background_task: bool = False,
    ) -> Union[Any, Response[T]]:
        return self.call(
            verb="POST",
            operation=operation,
            payload=payload,
            file=file,
            expect=expect,
            asynchronous=asynchronous,
            debug=debug,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
            raw_response=raw_response,
            is_app_call=app_call,
            app_owner=app_owner,
            app_id=app_id,
            app_instance_id=app_instance_id,
            as_background_task=as_background_task,
        )

    def get(
        self,
        operation: str,
        payload: Union[Request, dict] = None,
        file: Any = None,
        expect: Any = None,
        asynchronous: bool = False,
        debug: bool = False,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
        raw_response: bool = False,
        app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,
        as_background_task: bool = False,
    ) -> Union[Any, Response[T]]:
        return self.call(
            verb="GET",
            operation=operation,
            payload=payload,
            file=file,
            expect=expect,
            asynchronous=asynchronous,
            debug=debug,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
            raw_response=raw_response,
            is_app_call=app_call,
            app_owner=app_owner,
            app_id=app_id,
            app_instance_id=app_instance_id,
            as_background_task=as_background_task,
        )



from __future__ import annotations

import logging
from typing import Any, Dict, List

from steamship import Block, Configuration, PluginInstance, SteamshipError
from steamship.base import Client, Response
from steamship.base.base import IResponse
from steamship.base.tasks import TaskComment, TaskCommentList
from steamship.data import File
from steamship.data.app_instance import AppInstance
from steamship.data.embeddings import EmbedAndSearchRequest, EmbeddingIndex, QueryResults
from steamship.data.operations.tagger import TagRequest, TagResponse
from steamship.data.space import Space

_logger = logging.getLogger(__name__)


class Steamship(Client):
    """Steamship Python Client."""

    def __init__(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        web_base: str = None,
        space_id: str = None,
        space_handle: str = None,
        create_space: bool = False,
        profile: str = None,
        config_file: str = None,
        config: Configuration = None,
        **kwargs,
    ):
        super().__init__(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            web_base=web_base,
            space_id=space_id,
            space_handle=space_handle,
            create_space=create_space,
            profile=profile,
            config_file=config_file,
            config=config,
        )

    def create_index(
        self,
        handle: str = None,
        plugin_instance: str = None,
        upsert: bool = True,
        external_id: str = None,
        external_type: str = None,
        metadata: Any = None,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[EmbeddingIndex]:
        return EmbeddingIndex.create(
            client=self,
            handle=handle,
            plugin_instance=plugin_instance,
            upsert=upsert,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def upload(
        self,
        filename: str = None,
        content: str = None,
        mime_type: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[File]:
        return File.create(
            self,
            filename=filename,
            content=content,
            mime_type=mime_type,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def embed_and_search(
        self,
        query: str,
        docs: List[str],
        plugin_instance: str,
        k: int = 1,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[QueryResults]:
        req = EmbedAndSearchRequest(query=query, docs=docs, plugin_instance=plugin_instance, k=k)
        return self.post(
            "plugin/instance/embeddingSearch",
            req,
            expect=QueryResults,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def use(
        self,
        app_handle: str,
        handle: str = None,
        config: Dict[str, Any] = None,
        version: str = None,
        reuse: bool = True,
    ) -> AppInstance:
        """Creates or loads an instance named `handle` of App named `app_handle`."""
        instance = AppInstance.create(
            self,
            app_handle=app_handle,
            app_version_handle=version,
            handle=handle,
            config=config,
            upsert=reuse,
        )

        if instance.error:
            raise instance.error
        if not instance.data:
            raise SteamshipError(
                f"Unable to create an instance of App {app_handle} with handle {handle}."
            )
        return instance.data

    def use_plugin(
        self,
        plugin_handle: str,
        handle: str = None,
        config: Dict[str, Any] = None,
        version: str = None,
        reuse: bool = True,
    ) -> PluginInstance:
        """Creates or loads an instance named `handle` of a Plugin named `plugin_handle`."""
        instance = PluginInstance.create(
            self,
            plugin_handle=plugin_handle,
            plugin_version_handle=version,
            handle=handle,
            config=config,
            upsert=reuse,
        )

        if instance.error:
            raise instance.error
        if not instance.data:
            raise SteamshipError(
                f"Unable to create an instance of Plugin {plugin_handle} with handle {handle}."
            )
        return instance.data

    def tag(
        self,
        doc: str,
        plugin_instance: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[TagResponse]:
        req = TagRequest(
            type="inline",
            file=File.CreateRequest(blocks=[Block.CreateRequest(text=doc)]),
            plugin_instance=plugin_instance,
        )
        return self.post(
            "plugin/instance/tag",
            req,
            expect=TagResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def for_space(self, space_id: str = None, space_handle: str = None) -> Steamship:
        """Returns a new Steamship client anchored in the provided space as its default.

        Providing either `space_id` or `space_handle` will work; both need not be provided.
        """
        client = Steamship()
        client.config = self.config.for_space(space_id=space_id, space_handle=space_handle)
        return client

    def get_space(self) -> Space:
        # We should probably add a hard-coded way to get this. The client in a Steamship Plugin/App comes
        # pre-configured with an API key and the Space in which this client should be operating.
        # This is a way to load the model object for that space.
        logging.info("New client get_space")
        space = Space.get(self, id_=self.config.space_id, handle=self.config.space_handle)
        if not space.data:
            logging.error("Unable to get space.")
            raise SteamshipError(
                message="Error while retrieving the Space associated with this client config.",
                internal_message=f"space_id={self.config.space_id}   space_handle={self.config.space_handle}",
            )
        logging.info(f"Got space: {space.data.id}")
        return space.data

    def list_comments(
        self,
        task_id: str = None,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
    ) -> IResponse[TaskCommentList]:
        return TaskComment.list(
            client=self,
            task_id=task_id,
            external_id=external_id,
            external_type=external_type,
            external_group=external_group,
        )



from __future__ import annotations

from typing import Any, Dict, Type

from pydantic import BaseModel

from steamship.app import Response
from steamship.base import Client
from steamship.base.configuration import CamelModel


class User(CamelModel):
    client: Client = None
    id: str = None
    handle: str = None

    def dict(self, **kwargs) -> Dict[str, Any]:
        if "exclude" in kwargs:
            kwargs["exclude"] = {*(kwargs.get("exclude", set()) or set()), "client"}
        else:
            kwargs = {
                **kwargs,
                "exclude": {
                    "client",
                },
            }
        return super().dict(**kwargs)

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["user"] if "user" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def current(client: Client) -> Response[User]:
        return client.get("account/current", expect=User)



from steamship_tests import APPS_PATH
from steamship_tests.utils.deployables import deploy_app
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import AppInstance


def test_configurable_instance_invoke():
    greeting1 = "Hola"
    config_template = {"greeting": {"type": "string"}}
    instance_config = {"greeting": greeting1}
    client = get_steamship_client()
    hello_world_path = APPS_PATH / "configurable_hello_world.py"

    with deploy_app(
        client,
        hello_world_path,
        version_config_template=config_template,
        instance_config=instance_config,
    ) as (app, version, instance):
        # Now let's invoke it!
        # Note: we're invoking the data at configurable_hello_world.py in the steamship_tests/demo_apps folder
        res = instance.post("greet").data
        assert res == f"{greeting1}, Person"

        greeting2 = "Hallo"
        instance_config2 = {"greeting": greeting2}
        instance2 = AppInstance.create(
            client, app_id=app.id, app_version_id=version.id, config=instance_config2
        )
        instance2.wait()
        assert instance2.error is None
        assert instance2.data is not None
        instance2 = instance2.data

        res2 = instance2.post("greet").data
        assert res2 == f"{greeting2}, Person"

        # Test with quick-create
        greeting3 = "Howdy"
        instance_config3 = {"greeting": greeting3}
        instance3 = client.use(app.handle, f"{instance2.handle}-2", config=instance_config3)
        assert instance3.post("greet").data == f"{greeting3}, Person"
        assert instance3.id != instance2.id

        # Test that quick-create reuses an instance
        instance3a = client.use(app.handle, f"{instance2.handle}-2", config=instance_config3)
        assert instance3a.id == instance3.id

        # Test instance with quick create and no handle
        greeting4 = "How Do Dee"
        instance_config4 = {"greeting": greeting4}
        instance4 = client.use(app.handle, config=instance_config4)
        assert instance4.post("greet").data == f"{greeting4}, Person"
        assert instance4.id != instance3.id



from steamship.base.configuration import CamelModel


class Request(CamelModel):
    class Config:
        use_enum_values = True


class GetRequest(Request):
    id: str = None
    handle: str = None
    upsert: bool = None


class IdentifierRequest(Request):
    id: str = None
    handle: str = None



from __future__ import annotations

from typing import Any, Dict, Type

from pydantic import BaseModel

from steamship.base import Client, Request, Response
from steamship.base.configuration import CamelModel


class CreateAppVersionRequest(Request):
    app_id: str = None
    handle: str = None
    upsert: bool = None
    type: str = "file"
    config_template: Dict[str, Any] = None


class DeleteAppVersionRequest(Request):
    id: str


class ListPrivateAppVersionsRequest(Request):
    pass


class AppVersion(CamelModel):
    client: Client = None
    id: str = None
    app_id: str = None
    handle: str = None
    config_template: Dict[str, Any] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["appVersion"] if "appVersion" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        app_id: str = None,
        handle: str = None,
        filename: str = None,
        filebytes: bytes = None,
        upsert: bool = None,
        config_template: Dict[str, Any] = None,
    ) -> Response[AppVersion]:

        if filename is None and filebytes is None:
            raise Exception("Either filename or filebytes must be provided.")
        if filename is not None and filebytes is not None:
            raise Exception("Only either filename or filebytes should be provided.")

        if filename is not None:
            with open(filename, "rb") as f:
                filebytes = f.read()

        req = CreateAppVersionRequest(
            handle=handle, app_id=app_id, upsert=upsert, config_template=config_template
        )

        return client.post(
            "app/version/create",
            payload=req,
            file=("app.zip", filebytes, "multipart/form-data"),
            expect=AppVersion,
        )

    def delete(self) -> Response[AppVersion]:
        req = DeleteAppVersionRequest(id=self.id)
        return self.client.post("app/version/delete", payload=req, expect=AppVersion)



class TextTag:
    """This class provides a set of SpaCy-style constants to use with Tag(kind="token")."""

    TextWithWs = "textWithWs"
    Whitespace = "whitespace"
    Head = "head"
    LeftEdge = "leftEdge"
    RightEdge = "rightEdge"
    EntType = "entType"
    EntIob = "entIob"
    Lemma = "lemma"
    Normalized = "normalized"
    Shape = "shape"
    Prefix = "prefix"
    Suffix = "suffix"
    IsAlpha = "isAlpha"
    IsAscii = "isAscii"
    IsDigit = "isDigit"
    IsTitle = "isTitle"
    IsPunct = "isPunct"
    IsLeftPunct = "isLeftPunct"
    IsRightPunct = "isRightPunct"
    IsSpace = "isSpace"
    IsBracket = "isBracket"
    IsQuote = "isQuote"
    IsCurrency = "isCurrency"
    LikeUrl = "likeUrl"
    LikeNum = "likeNum"
    LikeEmail = "likeEmail"
    IsOov = "isOov"
    IsStop = "isStop"
    Lang = "lang"
    Embedding = "embedding"



"""CSV Blockifier - Steamship Plugin."""

import csv
import io
from typing import List, Optional, Type, Union

from pydantic import constr

from steamship.app import Response, create_handler
from steamship.base.error import SteamshipError
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags import Tag
from steamship.plugin.blockifier import Blockifier, Config
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.service import PluginRequest


class CsvBlockifier(Blockifier):
    """Converts CSV or TSV into Tagged Steamship Blocks.

    A CSV file is mapped onto 1 File
    Each row in the CSV file is mapped onto a block
    Each block has text which is extracted from a single column
    Each block has one or more tags extracted from the values in one ore more columns
    """

    class CsvBlockifierConfig(Config):
        text_column: str
        tag_columns: List[str]
        tag_kind: str = None
        delimiter: Optional[str] = ","
        quotechar: Optional[constr(max_length=1)] = '"'
        escapechar: Optional[constr(max_length=1)] = "\\"
        newline: Optional[str] = "\\n"
        skipinitialspace: Optional[bool] = False

        def __init__(self, **kwargs):
            if isinstance(kwargs["tag_columns"], str):
                kwargs["tag_columns"] = kwargs["tag_columns"].split(",")
            super().__init__(**kwargs)

    def config_cls(self) -> Type[CsvBlockifierConfig]:
        return self.CsvBlockifierConfig

    def _get_text(self, row) -> str:
        text = row.get(self.config.text_column)
        if text:
            text = text.replace(self.config.newline, "\n")
        return text

    def _get_tags(self, row) -> List[str]:
        tag_values = []
        for tag_column in self.config.tag_columns:
            tag_value = row.get(tag_column)
            if tag_value:
                tag_values.append(tag_value.replace(self.config.newline, "\n"))
        return tag_values

    def run(
        self, request: PluginRequest[RawDataPluginInput]
    ) -> Union[Response, Response[BlockAndTagPluginOutput]]:
        if request is None or request.data is None or request.data.data is None:
            return Response(
                error=SteamshipError(message="Missing data field on the incoming request.")
            )
        data = request.data.data  # TODO (enias): Simplify
        if isinstance(data, bytes):
            data = data.decode("utf-8")

        if not isinstance(data, str):
            return Response(
                error=SteamshipError(message="The incoming data was not of expected String type")
            )

        reader = csv.DictReader(
            io.StringIO(data),
            delimiter=self.config.delimiter,
            quotechar=self.config.quotechar,
            escapechar=self.config.escapechar,
            skipinitialspace=self.config.skipinitialspace,
        )
        file = File.CreateRequest(blocks=[])
        for row in reader:
            text = self._get_text(row)
            tag_values = self._get_tags(row)

            block = Block.CreateRequest(
                text=text,
                tags=[
                    Tag.CreateRequest(kind=self.config.tag_kind, name=tag_value)
                    for tag_value in tag_values
                ],
            )
            file.blocks.append(block)

        return Response(data=BlockAndTagPluginOutput(file=file))


handler = create_handler(CsvBlockifier)



import pytest
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_name

from steamship import Space, SteamshipError
from steamship.data.user import User


def test_connect():
    """Test basic connection"""
    client = get_steamship_client()
    assert client.config is not None
    assert client.config.profile == "test"
    assert client.config.api_key is not None


def test_user():
    client = get_steamship_client()
    user = User.current(client).data
    assert user.id is not None
    assert user.handle is not None


def test_client_creation_with_new_autogen_space():
    client1 = get_steamship_client(create_space=True)
    client2 = get_steamship_client()
    client3 = get_steamship_client()

    # Even though clients 2 and 3 are using the default space, the client constructor will contact the
    # Engine to load its ID and handle so that we always actively maintain an ID and Handle locally.
    assert client2.config.space_id is not None
    assert client2.config.space_handle is not None
    assert client3.config.space_id is not None
    assert client3.config.space_handle is not None

    assert client1.config.space_id is not None
    assert client1.config.space_handle is not None

    Space(client=client1, id=client1.config.space_id).delete()


def test_client_creation_space_failure_modes():
    # Can't create a space by specifying an ID
    with pytest.raises(SteamshipError):
        get_steamship_client(space_id="foo", create_space=True)

    client1 = get_steamship_client(create_space=True)

    # Can't create a space with a handle that already exists
    with pytest.raises(SteamshipError):
        get_steamship_client(space_handle=client1.config.space_handle, create_space=True)

    Space(client=client1, id=client1.config.space_id).delete()


def test_client_creation_new_space_custom_handle():
    name = random_name().lower()
    client1 = get_steamship_client(space_handle=name, create_space=True)

    assert client1.config.space_id is not None
    assert client1.config.space_handle == name

    Space(client=client1, id=client1.config.space_id).delete()


def test_client_creation_recall_existing_space():
    client1 = get_steamship_client(create_space=True)
    client2 = get_steamship_client(space_id=client1.config.space_id)
    client3 = get_steamship_client(space_handle=client1.config.space_handle)

    assert client1.config.space_id == client2.config.space_id
    assert client1.config.space_id == client3.config.space_id
    assert client1.config.space_handle == client2.config.space_handle
    assert client1.config.space_handle == client3.config.space_handle

    Space(client=client1, id=client1.config.space_id).delete()


def test_default_space_loaded():
    """Tests that the client actively loads the default space ID and Handle, and that we can revert to it later."""
    client1 = get_steamship_client(create_space=True)
    client2 = get_steamship_client(space_id=client1.config.space_id)
    client_default = get_steamship_client()

    assert client1.config.space_id == client2.config.space_id
    assert client1.config.space_id != client_default.config.space_id
    assert client1.config.space_handle == client2.config.space_handle
    assert client1.config.space_handle != client_default.config.space_handle
    assert client_default.config.space_handle == "default"

    client1.switch_space()
    assert client1.config.space_id != client2.config.space_id
    assert client1.config.space_id == client_default.config.space_id
    assert client1.config.space_handle != client2.config.space_handle
    assert client1.config.space_handle == client_default.config.space_handle

    # client2 has the newly created space.
    Space(client=client2, id=client2.config.space_id).delete()



#
# This is the CLIENT-side abstraction for an app.
#
# If you are implementing an app, see: steamship.app.server.App
#

from __future__ import annotations

from typing import Any, Type

from pydantic import BaseModel

from steamship.base import Client, Request, Response


class CreateAppRequest(Request):
    id: str = None
    handle: str = None
    upsert: bool = None


class DeleteAppRequest(Request):
    id: str


class ListPrivateAppsRequest(Request):
    pass


class GetAppRequest(Request):
    id: str = None
    handle: str = None


class App(BaseModel):
    client: Client = None
    id: str = None
    handle: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["app"] if "app" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(client: Client, handle: str = None, upsert: bool = None) -> Response[App]:
        req = CreateAppRequest(handle=handle, upsert=upsert)
        return client.post("app/create", payload=req, expect=App)

    @staticmethod
    def get(client: Client, handle: str):
        return client.post("app/get", GetAppRequest(handle=handle), expect=App)

    def delete(self) -> Response[App]:
        req = DeleteAppRequest(id=self.id)
        return self.client.post("app/delete", payload=req, expect=App)



from __future__ import annotations

from steamship.base.configuration import CamelModel
from steamship.data.file import File


class BlockAndTagPluginOutput(CamelModel):
    file: File.CreateRequest = None



from __future__ import annotations

from typing import Any, Dict, List, Optional, Set, Type, TypeVar

from pydantic import BaseModel

from steamship.base.base import IResponse
from steamship.base.configuration import CamelModel
from steamship.base.metadata import metadata_to_str, str_to_metadata
from steamship.base.request import Request

T = TypeVar("T")


class CreateTaskCommentRequest(Request):
    task_id: str
    external_id: str = None
    external_type: str = None
    external_group: str = None
    metadata: str = None
    upsert: bool = None


class ListTaskCommentRequest(Request):
    task_id: str = None
    external_id: str = None
    external_type: str = None
    external_group: str = None


class DeleteTaskCommentRequest(Request):
    id: str = None


class TaskComment(CamelModel):
    client: Any = None
    id: str = None
    user_id: str = None
    task_id: str = None
    external_id: str = None
    external_type: str = None
    external_group: str = None
    metadata: Any = None
    created_at: str = None

    def __init__(self, **kwargs):
        kwargs["metadata"] = str_to_metadata(kwargs.get("metadata"))
        super().__init__(**kwargs)

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["taskComment"] if "taskComment" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Any,  # TODO (Enias): Solve circular dependency
        task_id: str = None,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
        metadata: Any = None,
        upsert: bool = True,
    ) -> IResponse[TaskComment]:
        req = CreateTaskCommentRequest(
            taskId=task_id,
            external_id=external_id,
            external_type=external_type,
            externalGroup=external_group,
            metadata=metadata_to_str(metadata),
            upsert=upsert,
        )
        return client.post(
            "task/comment/create",
            req,
            expect=TaskComment,
        )

    @staticmethod
    def list(
        client: Any,
        task_id: str = None,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
    ) -> IResponse[TaskCommentList]:
        req = ListTaskCommentRequest(
            taskId=task_id,
            external_id=external_id,
            external_type=external_type,
            externalGroup=external_group,
        )
        return client.post(
            "task/comment/list",
            req,
            expect=TaskCommentList,
        )

    def delete(self) -> IResponse[TaskComment]:
        req = DeleteTaskCommentRequest(id=self.id)
        return self.client.post(
            "task/comment/delete",
            req,
            expect=TaskComment,
        )


class TaskCommentList(CamelModel):
    # TODO (enias): Not needed
    comments: List[TaskComment]


class TaskState:
    waiting = "waiting"
    running = "running"
    succeeded = "succeeded"
    failed = "failed"


class TaskType:
    internal_api = "internalApi"
    train = "train"
    infer = "infer"


class TaskRunRequest(Request):
    task_id: str


class TaskStatusRequest(Request):
    task_id: str


class Task(CamelModel):
    """Encapsulates a unit of asynchronously performed work."""

    client: Any = None  # Steamship client

    task_id: str = None  # The id of this task
    user_id: str = None  # The user who requested this task
    space_id: str = None  # The space in which this task is executing

    input: str = None  # The input provided to the task
    output: str = None  # The output of the task
    state: str = None  # A value in class TaskState

    status_message: str = None  # User-facing message concerning task status
    status_suggestion: str = None  # User-facing suggestion concerning error remediation
    status_code: str = None  # User-facing error code for support assistance
    status_created_on: str = None  # When the status fields were last set

    task_type: str = None  # A value in class TaskType; for internal routing
    task_executor: str = None  #
    task_created_on: str = None  # When the task object was created
    task_last_modified_on: str = None  # When the task object was last modified

    # Long Running Plugin Support
    # The `remote_status_*` fields govern how Steamship Plugins can communicate long-running work back to the engine.
    # If instead of sending data, the plugin sends a status with these fields set, the engine will begin polling for
    # updates, echoing the contents of these fields back to the plugin to communicate, e.g., the jobId of the work
    # being checked. When the work is complete, simply respond with the Response `data` field set as per usual.
    remote_status_input: Optional[
        Dict
    ] = None  # For re-hydrating state in order to check remote status.
    remote_status_output: Optional[
        Dict
    ] = None  # For reporting structured JSON state for error diagnostics.
    remote_status_message: str = None  # User facing message string to report on remote status.

    assigned_worker: str = None  # The worker assigned to complete this task
    started_at: str = None  # When the work on this task began

    max_retries: int = None  # The maximum number of retries allowed for this task
    retries: int = None  # The number of retries already used.

    def dict(self, **kwargs) -> Dict[str, Any]:
        if "exclude" in kwargs:
            kwargs["exclude"] = {*(kwargs.get("exclude", set()) or set()), "client"}
        else:
            kwargs = {
                **kwargs,
                "exclude": {
                    "client",
                },
            }
        return super().dict(**kwargs)

    def update(self, other: Optional[Task] = None):
        """Incorporates a `Task` into this object."""
        other = other or Task()
        for k, v in other.__dict__.items():
            self.__dict__[k] = v

    def add_comment(
        self,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
        metadata: Any = None,
        upsert: bool = True,
    ) -> IResponse[TaskComment]:
        return TaskComment.create(
            client=self.client,
            task_id=self.task_id,
            external_id=external_id,
            external_type=external_type,
            external_group=external_group,
            metadata=metadata,
            upsert=upsert,
        )

    def list_comments(self) -> IResponse[TaskCommentList]:
        return TaskComment.list(client=self.client, task_id=self.task_id)

    def post_update(self, fields: Set[str] = None) -> IResponse[Task]:
        """Updates this task in the Steamship Engine."""
        if not isinstance(fields, set):
            raise RuntimeError(f'Unexpected type of "fields": {type(fields)}. Expected type set.')
        body = self.dict(by_alias=True, include={*fields, "task_id"})
        return self.client.post("task/update", body, expect=Task)



from __future__ import annotations

from steamship.base import Request, Response

from ..file import File


class TagRequest(Request):
    type: str = None
    id: str = None
    name: str = None
    handle: str = None
    plugin_instance: str = None
    file: File.CreateRequest = None


class TagResponse(Response):
    file: File = None



from __future__ import annotations

from typing import Any, Dict, List

from steamship import File
from steamship.base import Client, Request, Response


class CorpusImportRequest(Request):
    # The Corpus Identifiers
    id: str = None
    handle: str = None
    type: str = "file"

    # Data for the plugin
    value: str = None
    data: str = None
    url: str = None
    plugin_instance: str = None
    file_importer_plugin_instance: str = None

    def dict(self, **kwargs) -> Dict[str, Any]:
        # TODO (enias): This can probably be bumped up the inheritance chain
        if "exclude" in kwargs:
            kwargs["exclude"] = {*(kwargs.get("exclude", set()) or set()), "client"}
        else:
            kwargs = {
                **kwargs,
                "exclude": {
                    "client",
                },
            }
        return super().dict(**kwargs)


class CorpusImportResponse(Response):
    client: Client = None
    file_import_requests: List[File.CreateRequest] = None



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Type

from steamship.app import post
from steamship.app.response import Response
from steamship.base import Client
from steamship.plugin.config import Config
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput
from steamship.plugin.service import PluginRequest, PluginService, TrainablePluginService
from steamship.plugin.trainable_model import TrainableModel


# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a Tagger.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.tagger instead
# of this file.
#
class Tagger(PluginService[BlockAndTagPluginInput, BlockAndTagPluginOutput], ABC):
    # noinspection PyUnusedLocal
    def __init__(self, client: Client = None, config: Dict[str, Any] = None):
        super().__init__(client, config)
        if config:
            self.config = self.config_cls()(**config)

    @abstractmethod
    def config_cls(self) -> Type[Config]:
        raise NotImplementedError()

    @abstractmethod
    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> Response[BlockAndTagPluginOutput]:
        raise NotImplementedError()

    @post("tag")
    def run_endpoint(self, **kwargs) -> Response[BlockAndTagPluginOutput]:
        """Exposes the Tagger's `run` operation to the Steamship Engine via the expected HTTP path POST /tag"""
        return self.run(PluginRequest[BlockAndTagPluginInput].parse_obj(kwargs))


class TrainableTagger(TrainablePluginService[BlockAndTagPluginInput, BlockAndTagPluginOutput], ABC):
    # noinspection PyUnusedLocal
    def __init__(self, client: Client = None, config: Dict[str, Any] = None):
        super().__init__(client=client, config=config)
        if config:
            self.config = self.config_cls()(**config)

    @abstractmethod
    def config_cls(self) -> Type[Config]:
        raise NotImplementedError()

    @abstractmethod
    def run_with_model(
        self, request: PluginRequest[BlockAndTagPluginInput], model: TrainableModel
    ) -> Response[BlockAndTagPluginOutput]:
        raise NotImplementedError()

    # noinspection PyUnusedLocal
    @post("tag")
    def run_endpoint(self, **kwargs) -> Response[BlockAndTagPluginOutput]:
        """Exposes the Tagger's `run` operation to the Steamship Engine via the expected HTTP path POST /tag"""
        return self.run(PluginRequest[BlockAndTagPluginInput].parse_obj(kwargs))

    # noinspection PyUnusedLocal
    @post("getTrainingParameters")
    def get_training_parameters_endpoint(self, **kwargs) -> Response[TrainingParameterPluginOutput]:
        """Exposes the Service's `get_training_parameters` operation to the Steamship Engine via the expected HTTP path POST /getTrainingParameters"""
        return self.get_training_parameters(PluginRequest[TrainingParameterPluginInput](**kwargs))

    # noinspection PyUnusedLocal
    @post("train")
    def train_endpoint(self, **kwargs) -> Response[TrainPluginOutput]:
        """Exposes the Service's `train` operation to the Steamship Engine via the expected HTTP path POST /train"""
        logging.info(f"Tagger:train_endpoint called. Calling train {kwargs}")
        arg = PluginRequest[TrainPluginInput].parse_obj(kwargs)
        model = self.model_cls()()
        model.receive_config(config=self.config)

        if arg.is_status_check:
            return self.train_status(arg, model)
        else:
            return self.train(arg, model)
